// gDirRec.cpp -- Version 1.2

#include "gXDir.h"
#include "gDirRec.h"
#include "gcalendar.h"

#define FMODE_OCTAL_MASK 007777

int gxdirSort=1;
////////////////////////////////////////////////////////////
t_uint64 gdr_calc_block_size (t_uint64 v, sDuOpt& duOpt)
{
 // 'stat' will give number of blocks in 512octet units
 t_uint64 result = v / 2 + (v==0 || (v%2)!=0);
 return result;
}

t_uint64 gdr_calc_byte_size (t_uint64 v, t_uint64 blocksInFile, sDuOpt& duOpt)
{
 t_uint64 result;
 if ( duOpt.doExact ) return v;
#ifdef gDOS_SPEC
 if ( blocksInFile==0 ) {
     // stat returned a poor result (like with DevC++ default libs): assuming 4096-octet blocks
     blocksInFile = (v / (t_uint64)4096) + ((v % 4096)!=0 || v==0);
     return blocksInFile*4096;
 }
#endif //gDOS_SPEC
 result = blocksInFile*512;
 return result;
}

t_uint64 gdr_calc_aprox_unit_size (t_uint64 byteSize,
				   t_uint64 divisor)
{
 ASSERTION(divisor>0,"divisor>0");
 return (byteSize / divisor) + (byteSize==0 || (byteSize%divisor)!=0);
}
////////////////////////////////////////////////////////////
// External additional methods
////////////////////////////////////////////////////////////
int gdr_strip_name (int stripLevel,
		    int fallback,
		    char slashChr,
		    gString& sName)
{
 int result( 0 );
 unsigned pos;
 gString sFallbackName( sName );

 if ( slashChr==0 ) slashChr = gSLASHCHR;

 if ( stripLevel<0 ) {
     if ( sFallbackName[ 1 ]=='.' ) {
	 sFallbackName.Delete( 1, 1 );
	 pos = sFallbackName[ 1 ]==slashChr;
	 if ( pos ) {
	     sFallbackName.Delete( 1, 1 );
	     if ( sFallbackName.IsEmpty()==false ) {
		 sName = sFallbackName;
	     }
	 }
     }
 }

 if ( stripLevel<=0 ) return 0;

 for ( ; stripLevel>0; stripLevel--) {
     for ( ; sName[ 1 ]==slashChr; ) {
	 sName.Delete( 1, 1 );
     }
     pos = sName.Find( slashChr );
     sName.Delete( 1, pos );
     result--;
 }

 sName.Trim();
 if ( sName.IsEmpty()==false )
     return 0;

 sName = sFallbackName;
 return result;
}

char* gdr_strip_name_str (char* str, int stripLevel, gString& sName)
{
 if ( str==nil )
     return nil;
 sName.Set( str );
 gdr_strip_name( stripLevel, 1, '\0', sName );
 return sName.Str();
}
////////////////////////////////////////////////////////////
// Main External methods
////////////////////////////////////////////////////////////
int gdr_du_dir (char* str,
		FILE* fOut,
		FILE* fRepErr,
		gHashTriple& myHash,
		sDuOpt& duOpt,
		int verboseLevel,
		t_uint64& byteSize,
		t_uint64& blockSize,
		int stripLevel,
		int& recControl,
		int& nErrors)
{
 int thisError;
 unsigned k, n;
 bool isDir;
 gKey hKey( gKey::e_UInt64 );
 gElem* ptrElem;
 gElem* ptrUsed;

 ASSERTION(fRepErr,"fRepErr");
 if ( duOpt.aMaxDepth>-1 && recControl>=duOpt.aMaxDepth ) {
     if ( verboseLevel>=3 ) {
	 fprintf(fRepErr,"Depth level, skipping: %s\n",str);
     }
     return 0;
 }

 if ( duOpt.doExact ) {
     blockSize = 0;
     byteSize = 0;
 }
 else {
     blockSize = 1;  // One dir occupies only one minor(512b) block
     byteSize = gdr_calc_byte_size( 0, blockSize, duOpt );
 }

 gXDir aDir( str );
 thisError = aDir.lastOpError;
 if ( thisError!=0 ) {
     fprintf(fRepErr,"Cannot access directory (%s): %s\n",
	     gFileControl::Self().ErrorStr( thisError ),
	     str);
     nErrors++;
     return 2;
 }
 else {
     if ( aDir.ioFails ) {
	 fprintf(fRepErr,"Failed to stat: %d element(s)\n",aDir.ioFails);
	 for (unsigned iter=1; iter<=aDir.errUnstatL.N(); iter++) {
	     fprintf(fRepErr,"!\t%s\n",aDir.errUnstatL.Str(iter));
	 }
     }
 }

 if ( recControl>9999 ) {
     fprintf(fRepErr,"Too many nested dirs... (at least %d)\n",recControl);
     return -1;
 }
 recControl++;

 bool isHardLink;
 t_inode inode;
 t_uint16 nLinks;
 char* aName( nil );
 gFileSysName* fSysPtr;
 gFileStat* pStat;
 gStorage* myObj( nil );

 n = aDir.N();

 aDir.SortDir( gxdirSort = duOpt.sortCode );
 ASSERTION(aDir.SortedList(),"aDir.SortedList()");  // Currently only sorted result is supported, even if it is not used
 ptrUsed = aDir.StartSortedPtr();

 for (k=1, ptrElem=ptrUsed; k<=n; k++, ptrElem=ptrElem->next) {
     ASSERTION(ptrElem,"ptrElem");
     myObj = ptrElem->me;
     ASSERTION(myObj,"myObj");
     DBGPRINT_MIN("DBG: recControl=%d, %u/%u #%u: myObj: kind=%d: %s\n",
		  recControl,
		  k, n, aDir.SortedList()->N(),
		  myObj->Kind(),
		  myObj->Kind()==gStorage::e_StoreExtend?myObj->Str():nil);
     fSysPtr = ((gXFilePtr*)myObj)->SysName();
     pStat = fSysPtr->GetStatPtr();
     aName = fSysPtr->Str();
     gString sName( str );
     sName.Add( aName );
     aName = sName.Str();
     inode = pStat->status.inode;
     nLinks = pStat->status.nLinks;
     isDir = fSysPtr->IsDirectory();

     DBGPRINT_MIN("DBG: str='%s', aName='%s', isDir?%c, inode=%lu, nLinks=%u\n",
		  str,
		  aName,
		  ISyORn( isDir ),
		  (unsigned long)inode,
		  nLinks);

     // Check if the hard-link was already counted.
     // A stupid question: "which is the master file when having 'a' and 'b'?"
     // Answer is: the first that was created will appear in the count...
     isHardLink = isDir ? (nLinks>2) : (nLinks>1);

     if ( isHardLink ) {
	 hKey.Set( inode );
	 if ( duOpt.doCountHardLinks==false ) {
	     // ...then it's a hard-link and we do not count it
	     if ( myHash.FindKey( hKey ) ) continue;
	     myHash.AddTriple( hKey, aName, nil );
	 }
	 else {
	     if ( verboseLevel>=3 && duOpt.doCountHardLinks==true ) myHash.AddTriple( hKey, aName, nil );
	 }
     }

     t_uint64 aByteSize=0;
     t_uint64 aBlockSize=0;
     if ( isDir ) {
	 if ( verboseLevel>=3 ) fprintf(fRepErr,"Recursing (%d) into '%s'\n",recControl,aName);
	 // Recursive call to 'gdr_du_dir'
	 thisError = gdr_du_dir( aName, fOut, fRepErr, myHash, duOpt, verboseLevel, aByteSize, aBlockSize, stripLevel, recControl, nErrors );
	 if ( thisError!=0 ) {
	     nErrors++;
	     continue;
	 }

	 gdr_du_print( aName, fOut, *pStat, duOpt, verboseLevel, aByteSize, aBlockSize, isHardLink, stripLevel, false );
     }
     else {
	 t_uint32 blocksInFile;
	 aBlockSize = (t_uint64)(blocksInFile = pStat->statusL.blocks);
	 if ( pStat->IsLink() )
	     aByteSize = 0;
	 else
	     aByteSize = gdr_calc_byte_size( (t_uint64)(pStat->statusL.size), (t_uint64)blocksInFile, duOpt );

	 if ( duOpt.doShowAll ) {
	     gdr_du_print( aName, fOut, *pStat, duOpt, verboseLevel, aByteSize, aBlockSize, isHardLink, stripLevel, false );
	 }
     }
     blockSize += aBlockSize;
     byteSize += aByteSize;
 }//end FOR k

 recControl--;
 return 0;
}
////////////////////////////////////////////////////////////
int gdr_du_print (char* str,
		  FILE* fOut,
		  gFileStat& fStat,
		  sDuOpt& duOpt,
		  int verboseLevel,
		  t_uint64 byteSize,
		  t_uint64 blockSize,
		  bool isHardLink,
		  int stripLevel,
		  bool isArg)
{
 static char strBuf[ 128 ];
 char fmtStr[32];
 t_uint64 uVal64;
 t_uint64 megaDivisor( duOpt.doByMegaBytes ? 1024 : 1 );
 unsigned long long val;
 const int specialValue( 1000 );
 int zValue( duOpt.zValue );
 int calcValue( (zValue >= specialValue) ? zValue - specialValue : zValue );
 bool isVeryVerbose( verboseLevel>=3 );
 bool doShowHardLinkSymbol = isVeryVerbose && isHardLink;

 ASSERTION(str,"str");

 if ( duOpt.doSummarize ) {
     if ( isArg==false ) return 0;
 }

 // Strip path, if requested to:
 gString sName( str );
 if ( gdr_strip_name( stripLevel, 1, '\0', sName )<0 && isVeryVerbose==true ) {
     sName.InsertStr( "... " );
 }
 str = sName.Str();

 if ( duOpt.doByByteSize==false && duOpt.doByBlockSize==false ) {
     if ( zValue>=specialValue && fStat.IsDirectory()==true ) {
	 return 0;
     }
     fprintf(fOut,"%s\n",str);
     return 0;
 }
 uVal64 = byteSize;
 if ( duOpt.doByBlockSize ) {
#ifdef gDOS_SPEC
     // DevC++ does not support 'blockSize' properly:
     // just calc. the most probable number of blocks used (assuming adjacent disk blocks!)
     uVal64 = gdr_calc_aprox_unit_size( byteSize, 1024 );
#else
     uVal64 = gdr_calc_block_size( blockSize, duOpt );
#endif //gDOS_SPEC
 }
 else {
     uVal64 = gdr_calc_aprox_unit_size( uVal64, megaDivisor );
 }

 val = (unsigned long long)gdr_calc_aprox_unit_size( uVal64, megaDivisor );

 switch ( calcValue ) {
 case 15:
     snprintf( strBuf, sizeof(strBuf), "%lu,%lu,%lu,0x%04X\t",
	       fStat.status.aTime,
	       fStat.status.mTime,
	       fStat.status.cTime,
	       fStat.status.mode & 0xFFFF );
     break;
 case 1:
     snprintf( strBuf, sizeof(strBuf), "%lu\t",
	       fStat.status.aTime );
     break;
 case 2:
     snprintf( strBuf, sizeof(strBuf), "%lu\t",
	       fStat.status.mTime );
     break;
 case 4:
     snprintf( strBuf, sizeof(strBuf), "%lu\t",
	       fStat.status.cTime );
     break;
 case 8:
     snprintf( strBuf, sizeof(strBuf), "0x%04X\t",
	       fStat.status.mode & 0xFFFF );
     break;
 case 10:
     snprintf( strBuf, sizeof(strBuf), "%lu,%04o\t",
	       fStat.status.mTime,
	       fStat.status.mode & FMODE_OCTAL_MASK );
     break;
 case 91:
 case 191:  // Like 91, but with local-time
     if ( duOpt.viewMask ) {
	 t_stamp fileStamp( fStat.status.mTime );
	 gDateTime modifiedDtTm( fileStamp );

#ifdef GNORM_USE_LOCALTIME
	 if ( zValue==191 ) {
	     // Use local time instead of gmtime
	     struct tm* pTM;
	     time_t localStamp( (time_t)fileStamp );
	     pTM = localtime( &localStamp );
	     if ( pTM ) {
		 modifiedDtTm.year = (t_int16)pTM->tm_year+1900;
		 modifiedDtTm.month = (t_uint8)pTM->tm_mon+1;
		 modifiedDtTm.day = (t_uint8)pTM->tm_mday;
		 modifiedDtTm.hour = (t_uint8)pTM->tm_hour;
		 modifiedDtTm.minu = (t_uint8)pTM->tm_min;
		 modifiedDtTm.sec =(t_uint8)pTM->tm_sec;
		 modifiedDtTm.wday = (t_uint8)pTM->tm_wday;
		 modifiedDtTm.yday = (t_uint16)pTM->tm_yday;
		 modifiedDtTm.isdst = (t_int8)pTM->tm_isdst;
	     }
	 }
#endif //GNORM_USE_LOCALTIME
	 if ( duOpt.doSupportInode ) {
	     snprintf( strBuf, sizeof(strBuf), "%04u-%02u-%02u %02u:%02u:%02u,%04o,%lu\t",
		       modifiedDtTm.year,
		       modifiedDtTm.month,
		       modifiedDtTm.day,
		       modifiedDtTm.hour,
		       modifiedDtTm.minu,
		       modifiedDtTm.sec,
		       fStat.status.mode & FMODE_OCTAL_MASK,
		       (unsigned long)fStat.status.inode);
	 }
	 else {
	     snprintf( strBuf, sizeof(strBuf), "%04u-%02u-%02u %02u:%02u:%02u,%04o,-\t",
		       modifiedDtTm.year,
		       modifiedDtTm.month,
		       modifiedDtTm.day,
		       modifiedDtTm.hour,
		       modifiedDtTm.minu,
		       modifiedDtTm.sec,
		       fStat.status.mode & FMODE_OCTAL_MASK);
	 }
     }
     else {
	 if ( duOpt.doSupportInode ) {
	     snprintf( strBuf, sizeof(strBuf), "%lu,%04o,%lu\t",
		       fStat.status.mTime,
		       fStat.status.mode & FMODE_OCTAL_MASK,
		       (unsigned long)fStat.status.inode);
	 }
	 else {
	     snprintf( strBuf, sizeof(strBuf), "%lu,%04o,-\t",
		       fStat.status.mTime,
		       fStat.status.mode & FMODE_OCTAL_MASK);
	 }
     }
     break;
 case 0:
 default:
     strBuf[ 0 ] = 0;
     break;
 }

 sprintf(fmtStr,duOpt.doSupportLL?"%s%%%s%ullu\t%%s%%s\n":"%s%%%s%ulu\t%%s%%s\n",
	 doShowHardLinkSymbol?"@":"\0",
	 duOpt.sFmtBasicDigits.Str(),
	 duOpt.GetMaxDigits()-(unsigned)doShowHardLinkSymbol);
 DBGPRINT_MIN("DBG: LL?%c %d\t%s\n",
	      ISyORn(duOpt.doSupportLL),
	      zValue,
	      fmtStr);
 if ( zValue>=specialValue ) {
#ifdef gDOS_SPEC
     // Because .IsDirectory does not work!
     if ( fStat.status.mode & 0040000 ) {
	 return 0;
     }
#endif //gDOS_SPEC
     if ( fStat.IsDirectory()==true ) {
	 return 0;
     }
 }

 if ( duOpt.doSupportLL )
     fprintf(fOut,fmtStr,val,strBuf,str);
 else
     fprintf(fOut,fmtStr,(unsigned long)val,strBuf,str);

 // e.g.: printf("%-8llu%s\n",val,str) : DevC++ does not support printf long-long
 return 1;
}
////////////////////////////////////////////////////////////
int gdr_du_dir_list (char* str,
		     FILE* fOut,
		     FILE* fRepErr,
		     gHashTriple& myHash,
		     sDuOpt& duOpt,
		     int verboseLevel,
		     gList& lPaths,
		     t_uint64& byteSize,
		     t_uint64& blockSize,
		     int& recControl,
		     int& nErrors)
{
 // Sorry for almost copying 'gdr_du_dir' into here,
 // but I was terribly lazy, and 'gdr_du_dir' worked so well!

 int thisError;
 unsigned k, n;
 bool isDir;
 gKey hKey( gKey::e_UInt64 );

 ASSERTION(fRepErr,"fRepErr");
 if ( duOpt.aMaxDepth>-1 && recControl>=duOpt.aMaxDepth ) {
     if ( verboseLevel>=3 ) {
	 fprintf(fRepErr,"Depth level, skipping: %s\n",str);
     }
     return 0;
 }

 gXDir aDir( str );
 thisError = aDir.lastOpError;
 DBGPRINT_MIN("DBG: gdr_du_dir_list DIR, recControl=%d, code=%d,%d,ioFails=%d,%d: %s\n",
	      recControl,
	      thisError,
	      (int)aDir.errUnstatL.N(),
	      aDir.ioFails,
	      aDir.IsOk()==false,
	      str);
 if ( thisError!=0 ) {
     fprintf(fRepErr,"Cannot access directory (%s): %s\n",
	     gFileControl::Self().ErrorStr( thisError ),
	     str);
     nErrors++;
     return 2;
 }

 if ( recControl>9999 ) {
     fprintf(fRepErr,"Too many nested dirs... (at least %d)\n",recControl);
     return -1;
 }
 recControl++;


 bool isHardLink;
 t_inode inode;
 t_uint16 nLinks;
 char* aName( nil );
 n = aDir.N();
 // error will count number of errors

 for (k=1; k<=n; k++) {
     gFileSysName* fSysPtr = (gFileSysName*)aDir.GetObjectPtr( k );
     gFileStat* pStat = fSysPtr->GetStatPtr();
     aName = fSysPtr->Str();
     gString sName( str );
     sName.Add( aName );
     aName = sName.Str();
     inode = pStat->status.inode;
     nLinks = pStat->status.nLinks;
     isDir = fSysPtr->IsDirectory();

     // Check if the hard-link was already counted.
     // A stupid question: "which is the master file when having 'a' and 'b'?"
     // Answer is: the first that was created will appear in the count...
     isHardLink = isDir ? (nLinks>2) : (nLinks>1);

     if ( isHardLink ) {
	 hKey.Set( inode );
	 if ( duOpt.doCountHardLinks==false ) {
	     // ...then it's a hard-link and we do not count it
	     if ( myHash.FindKey( hKey ) ) continue;
	     myHash.AddTriple( hKey, aName, nil );
	 }
	 else {
	     if ( verboseLevel>=3 && duOpt.doCountHardLinks==true ) myHash.AddTriple( hKey, aName, nil );
	 }
     }

     t_uint64 aByteSize=0;
     t_uint64 aBlockSize=0;
     if ( isDir ) {
	 if ( verboseLevel>=3 ) fprintf(fRepErr,"Recursing (%d) into '%s'\n",recControl,aName);
	 // Recursive call to 'gdr_du_dir<...>'
	 thisError = gdr_du_dir_list( aName, fOut, fRepErr, myHash, duOpt, verboseLevel, lPaths, aByteSize, aBlockSize, recControl, nErrors );
	 if ( thisError!=0 ) {
	     nErrors++;
	     continue;
	 }
	 lPaths.Add( aName );
     }
     else {
	 if ( pStat->IsLink()==false ) {
	     lPaths.Add( aName );
	     DBGPRINT_MIN("DBG: ADDED: %s\n",aName);
	 }
     }
 }//end FOR k

 recControl--;
 return 0;
}
////////////////////////////////////////////////////////////

