// gNorm.cpp -- Version 1.2


#define thisProgramVersion "Version 1.2"
#define thisProgramCopyright "Prized Season & Sons"
#define thisProgramYear 2009
#define thisProgramCopy "This is free software (GPL)\n\
There is no warranty, not even for MERCHANTABILITY or\n\
FITNESS FOR A PARTICULAR PURPOSE."

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "gNorm.h"
#include "garg.h"
#include "gstringext.h"
#include "ghash.h"

#include "gDirRec.h"
#include "gXDir.h"
#include "gNormList.h"
#include "gNormStubs.h"
#include "gSumLines.h"
#include "gNormIoHelp.h"
#include "gIoLists.h"

#ifdef gDOS_SPEC
#include <windows.h>
#endif //gDOS_SPEC
////////////////////////////////////////////////////////////

// Globals: I know they are ugly, but 'sum-a' is still experimental.
t_int32 globSumA_reqBehMask=SUMA_BEH_MASK_DEFAULT;

////////////////////////////////////////////////////////////
enum eSumAlgorithm {
    e_SumV,
    e_SumBSD,
    e_SumA
};
////////////////////////////////////////////////////////////
struct sOptNorm {
    sOptNorm ()
	: isVerbose( false ),
	  isVeryVerbose( false ),
	  isChecker( false ),
	  iValue( 0 ),
	  stripLevel( 0 ),
	  doShowAll( false ),
	  // 'ddb'
	  bs( 1 ),
	  count( 0 ) {
#ifdef gDOS_SPEC
	duOpt.doSupportInode = duOpt.doSupportLL = false;
#endif //gDOS_SPEC
    }
    ~sOptNorm () {
    }

    bool isVerbose, isVeryVerbose;
    bool isChecker;
    int iValue;
    int stripLevel;
    unsigned numMaxDigits;
    bool doShowAll;
    sDuOpt duOpt;
    gString sCheckerFile;
    gString sOutput;
    ;
    t_uint32 bs;
    t_uint32 count;

    // Specific hash table for classifying hard-links
    gHashTriple myHash;

    // Opt-ins...
    bool doSupportLL;  // Support print long-long
    bool doSupportInode;  // Support inode

    // Get methods
    bool DoCheck () {
	return isChecker;
    }
    bool DoShowHLinks () {
	return isVeryVerbose && duOpt.doCountHardLinks;
    }
    // .
};

////////////////////////////////////////////////////////////
// glib required functions
////////////////////////////////////////////////////////////
int gglobal_EventTerminate (int v)
{
 return 0;
}

////////////////////////////////////////////////////////////
int print_help (char* progStr)
{
 char
     *msgHelp = "%s - %s\n\n\
Usage:\n\
        %s command [--help] [OPTION] [NAME ...]\n\
\n\
Commands are:\n\
   check        Check files & dirs\n\
   du           Disk usage (like du)\n\
   ddb          Basic dd\n\
   touchb       Basic touch\n\
   stat         Basic stat command\n\
   sumv         Sum (System V sum algorithm)\n\
   suma         Sum alternate (faster) algorithm\n\
   pwd          Print name of current/working directory\n\
   sleep        Sleep for some seconds\n\
\n\
Options are:\n\
   -h           This help (or --help / --version)\n\
   -v           Verbose (use twice, more verbose)\n\
   -a           Show all files (or --all)\n\
   -c           Checker file (or --checker)\n\
   -o           Output to file instead of stdout (or --output)\n\
   -z N         Value option\n\
\n";

 printf(msgHelp,
	progStr,
	thisProgramVersion,
	progStr,
	progStr);
 return 0;
}

int print_help_command (char* progStr, char* cmdStr)
{
 int cmdNr=-1;
 char
     *msgHelpI[] = {
	 "%s : %s\n\
	Shows and checks all files recursively: especially relevant for root.\n\
	Returns 2 if there some of the items requested are not found,\n\
	or 1 if some of the files were not found.\n\
",
	 "%s : %s\n\
   -b           Show size in bytes (or --bytes)\n\
   -k           Show size in 1024-bytes units (--kilobytes)\n\
   -m           Show size in 1024*1024-bytes units (--megabytes)\n\
   -z N         Show type-mask: 1:aTime, 2:mTime, 4:cTime, 8:mode;\n\
                default is 0; popular values: 10 (mTime and mode);\n\
                191 shows mTime,OctalPermissions,inode;\n\
   --summarize  Show only total\n\
   --exact      Show the exact value, regardless of file-system\n\
   --no-sort    Do not sort alphabetically\n\
   --hard-links Show and count also hard-links\n\
   --max-depth=N  Show only if the directory is N or fewer levels below\n\
                  the argument\n\
   --strip=N      Strip first N paths from displayed dirs,\n\
                  N=-1 to strip leading ./;\n\
                  (e.g. --strip=1 shows 'def' instead of 'abc/def/file1')\n\
\n",
	 "%s : %s\n\
\tBasic dd\n\
\n\
Examples:\n\
	Output to file 'myout' 1024 octets ASCII 0 twice, ie. 2 Kbytes:\n\
		ddb --bs 1024 --count 2 #0 myout\n\
",
	 "%s : %s file [mTime [aTime]]\n\
\n\
Basic touch:\n\
	mTime is the modification time;\n\
	aTime is the access time;\n\
	if none is given, touches with current time.\n\
",
	 nil,
	 "%s : %s\n\
\tsystemV sum\n\
\n\
Examples:\n\
	Calculate sum of all files:\n\
		sum *\n\
	Calculate sum of input given:\n\
		echo -n abc | sum\n\
",
	 "%s : %s\n\
\tsum-A is an alternate (relaxed) sum\n\
\n\
   -a           Calculate sum for links\n\
   -z 1         Use an md5sum-like output\n\
\n\
Examples:\n\
	Calculate sum-A of all files:\n\
		suma *\n\
\n\
	Calculate sum-A of all files below our pwd:\n\
		gnorm suma `gnorm du -a *`\n\
	or\n\
		gnorm suma `gnorm du -a .`\n\
	or specify file names with stdin:\n\
		find * -type f | gnorm suma\n\
",
	 "%s : %s [--strip=N] [path ...[path]]\n\
\n\
Shows the current working directory,\n\
and optionally the paths.\n\
",
	 "%s : %s [-z=1] seconds [milliseconds [optional echoes]]\n\
\n\
Sleeps for n seconds (and optionally m miliseconds).\n\
If '-z=1' is entered, dots are shown, one per second.\n\
"
     };

 if ( strcmp(cmdStr,"check")==0 ) cmdNr=0;
 if ( strcmp(cmdStr,"du")==0 ) cmdNr=1;
 if ( strcmp(cmdStr,"ddb")==0 ) cmdNr=2;
 if ( strcmp(cmdStr,"touchb")==0 ) cmdNr=3;
 if ( strcmp(cmdStr,"sumv")==0 ) cmdNr=5;
 if ( strcmp(cmdStr,"suma")==0 ) cmdNr=6;
 if ( strcmp(cmdStr,"pwd")==0 ) cmdNr=7;
 if ( strcmp(cmdStr,"sleep")==0 ) cmdNr=8;
 if ( cmdNr<0 ) {
     printf("No help for command: %s\n",cmdStr);
     return -1;
 }

 printf(msgHelpI[cmdNr],
	progStr,
	cmdStr);
 return 0;
}
////////////////////////////////////////////////////////////
int print_version (char* progStr)
{
 char
     *msgVersion = "%s - %s\n\
\n\
Written by Henrique Moreira.\n\
\n\
Copyright (C) %u %s.\n\
%s\n";

 printf(msgVersion,
	progStr,
	thisProgramVersion,
	thisProgramYear,thisProgramCopyright,
	thisProgramCopy);
 return 0;
}
////////////////////////////////////////////////////////////
int do_stat_print (char* str, FILE* fOut, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error;
 long adjustStamp( (long)opt.iValue );
 bool isVerbose = opt.isVerbose;
 bool mostVerbose = opt.isVeryVerbose;
 t_inode inode;
 char* strErr;
 gFileStat probe( str );

 ASSERTION(fOut,"fOut");

 strErr = strerror( probe.lastOpError );
 // error = probe.Error()  <== not quite: a broken soft-link is still stat'able.
 error = probe.HasStat()==false;
 DBGPRINT("DBG: do_stat(%s), HasStat?%c, lastOpError=%d, error=%d\n",str,ISyORn(probe.HasStat()),probe.lastOpError,error);

 result = error;
 if ( error ) {
     if ( fRepErr ) fprintf(fRepErr,"%s: %s\n",str,strErr);
     return 1;
 }

 if ( mostVerbose )
     fprintf(fOut,"%s:\n",str);

 inode = probe.statusL.inode;

 fprintf(fOut,"Size: %lu, %lu\n",
	 (unsigned long)probe.status.size,
	 (unsigned long)probe.statusL.size);
 fprintf(fOut,"Blocks: %lu\n",
	 (unsigned long)probe.status.blocks);
 fprintf(fOut,"Mode: 0x%08lX%s",
	 (unsigned long)probe.status.mode,
	 mostVerbose ? " " : "\n" );
 if ( mostVerbose )
     fprintf(fOut,"(%04o)%s\n",
	     (unsigned)probe.status.mode & 007777,
	     (probe.status.mode & 006000) ? " (SetGID/UID)" : "\0");
 // mask above with 1777octal (0x3FF) would be enough,
 // but since sticky-bit ('t') is 1000octal, and
 // group setgid-bit is 2000octal (and 4000octal the user stuid-bit),
 // we mask additionally with 7000octal.
 fprintf(fOut,"Uid: %u, Gid: %u\n",
	 probe.status.uid,
	 probe.status.gid);
 fprintf(fOut,"#Links: %u\n",
	 (unsigned)probe.status.nLinks);
 fprintf(fOut,"Inode: %lu\n",
	 (unsigned long)inode);
 if ( isVerbose && probe.status.inode && probe.status.inode!=inode )
     fprintf(fOut,"Inode-to: %lu\n",
	     (unsigned long)probe.status.inode);
 fprintf(fOut,"Times%s: %lu, %lu, %lu\n",
	 isVerbose ? " (access/modify/change)" : "\0",
	 (unsigned long)((long)probe.status.aTime + adjustStamp),
	 (unsigned long)((long)probe.status.mTime + adjustStamp),
	 (unsigned long)((long)probe.status.cTime + adjustStamp));
 return 0;
}

int print_sum_x (FILE* fOut,
		 eSumAlgorithm alg,
		 int printMode,
		 bool hasError,
		 t_uint16 chkSum,
		 int stripLevel,
		 char* str)
{
 bool showSumThenFile( (printMode & 2)==0 );
 char* strName( sul_universal_set_checkline( str ) );
 const char* const strPrefix( alg==e_SumA ? SUMA_MAGIC_PREFIX : SUMV_MAGIC_PREFIX );

 if ( fOut==nil ) return -1;
 if ( printMode & 1 ) {
     if ( strName==nil ) return 1;

     gString sName( strName );
     gdr_strip_name( stripLevel, 1, '/', sName );
     strName = sName.Str();

     if ( hasError ) {
	 if ( showSumThenFile )
	     fprintf(fOut,"# [ERROR]\t%s\n",strName);
	 else
	     fprintf(fOut,"%s: [ERROR]\n",strName);
     }
     else {
	 if ( showSumThenFile )
	     fprintf(fOut,"%s%05u\t%s\n",strPrefix,chkSum,strName);
	 else
	     fprintf(fOut,"%s: %u\n",strName,chkSum);
     }
 }
 else {
     fprintf(fOut,"%u\n",chkSum);
 }
 return 0;
}
////////////////////////////////////////////////////////////
int do_dd_io (FILE* fIn, FILE* fOut, FILE* fRepErr,
	      FILE* fRepProgress,
	      gUCharBuffer& sBuf,
	      t_uint32 count)
{
 // Returns a positive error, a negative warning, or zero if ok.
 t_uint32 iter;
 size_t readElems=0;
 size_t wroteElems=0;

 ASSERTION(fOut,"fOut");
 DBGPRINT_MIN("DBG: fIn stdin?%c nil?%c count=%lu\n",
	      ISyORn(fIn==stdin),
	      ISyORn(fIn==nil),
	      (unsigned long)count);

 for (iter=1; fIn==stdin || count==0 || (count>0 && iter<=count); iter++) {
     if ( fIn ) {
	 readElems = fread( sBuf.uBuf, sBuf.size, 1, fIn );
	 if ( readElems<=0 ) break;
     }
     wroteElems = fwrite( sBuf.uBuf, sBuf.size, 1, fOut );
     DBGPRINT_MIN("DBG: iter %u/%u: readElems=%d wroteElems=%d\n",
		  iter, count,
		  readElems, wroteElems);
 }

 if ( wroteElems<=0 ) return -2;
 if ( readElems<=0 ) return fIn ? (fIn==stdin ? 0 : -1) : 0;
 return 0;
}
////////////////////////////////////////////////////////////
int do_du_print (char* str,
		 FILE* fOut,
		 gFileStat& fStat,
		 sOptNorm& opt,
		 t_uint64 byteSize,
		 t_uint64 blockSize,
		 bool isHardLink,
		 bool isArg)
{
 int verboseLevel( opt.isVeryVerbose ? 3 : (int)opt.isVerbose );
 int stripLevel( opt.stripLevel );
 int result( gdr_du_print( str, fOut, fStat, opt.duOpt, verboseLevel, byteSize, blockSize, isHardLink, stripLevel, isArg ) );
 DBGPRINT("DBG: gdr_du_print returned %d\n",result);
 return result;
}

int do_du_dir (char* str,
	       FILE* fOut,
	       FILE* fRepErr,
	       sOptNorm& opt,
	       t_uint64& byteSize,
	       t_uint64& blockSize,
	       int& recControl,
	       int& nErrors)
{
 int verboseLevel( opt.isVeryVerbose ? 3 : (int)opt.isVerbose );
 int stripLevel( opt.stripLevel );
 int result( gdr_du_dir( str, fOut, fRepErr, opt.myHash, opt.duOpt, verboseLevel, byteSize, blockSize, stripLevel, recControl, nErrors ) );
 DBGPRINT("DBG: gdr_du_dir returned %d\n",result);
 return result;
}

int do_du_dir_list (char* str,
		    FILE* fOut,
		    FILE* fRepErr,
		    sOptNorm& opt,
		    gNormList& lPaths,
		    t_uint64& byteSize,
		    t_uint64& blockSize,
		    int& recControl,
		    int& nErrors)
{
 int verboseLevel( opt.isVeryVerbose ? 3 : (int)opt.isVerbose );
 int result( gdr_du_dir_list( str, fOut, fRepErr, opt.myHash, opt.duOpt, verboseLevel, lPaths, byteSize, blockSize, recControl, nErrors ) );
 DBGPRINT("DBG: gdr_du_dir_list returned %d\n",result);
 return result;
}

int do_process__sysv_sum_file (char* str, t_uint16& chkSum, t_uint64& totalBytes)
{
 static t_uchar buf[8192];
 size_t i, bytes_read;
 FILE* fIn = str ? fopen( str, "rb" ) : stdin;

 unsigned s = 0;
 int r;
 int checksum;

 if ( fIn==nil ) return 2;
 int fd = fileno( fIn );
 ASSERTION(fd>=0,"fd>=0");

 totalBytes = 0;

 while (1) {
     bytes_read = read( fd, buf, sizeof(buf) );
     if ( bytes_read==0 ) break;
     if ( bytes_read<0 ) {
	 fclose( fIn );
	 return 1;
     }

     for (i=0; i<bytes_read; i++) s += buf[ i ];
     totalBytes += (t_uint64)bytes_read;
 }

 r = (s & 0xffff) + ((s & 0xffffffff) >> 16);
 checksum = (r & 0xffff) + (r >> 16);

 chkSum = (t_uint16)checksum;

 if ( fIn!=stdin ) fclose( fIn );
 return 0;
}

int do_process__suma_file (char* str,
			   FILE* fRepErr,
			   int verboseLevel,
			   t_uint16& chkSum,
			   t_uint64& totalBytes)
{
 static t_uchar buf[8192];
 FILE* fIn = str ? fopen( str, "rb" ) : nil;

 int error;
 int bogus( 0 );
 unsigned statBytesRead;
 off_t seeked;
 t_uint64 byteSizeCheck( totalBytes );
 gFileStat dummyStat;

 if ( fIn==nil ) return 2;
 int fd = fileno( fIn );
 ASSERTION(fIn!=stdin,"fIn!=stdin");  // Simply because we close the stream without checking if it's a stdin!

 if ( byteSizeCheck==0 ) {
     dummyStat.Update( fd );
     byteSizeCheck = (t_uint64)dummyStat.status.size;
     if ( totalBytes>0 && byteSizeCheck!=byteSizeCheck ) {
	 fclose( fIn );
	 return 3;  // Inconsistent size
     }
 }

 error = gio_stable_file( fd, fRepErr, 1, seeked );
 DBGPRINT_MIN("DBG: gio_stable_file returned %d, seeked=%lu, byteSizeCheck=%lu\n",
	      error,
	      (unsigned long)seeked,
	      (unsigned long)byteSizeCheck);

 switch ( error ) {
 case 0:
     totalBytes = 0;
     error = calc_sumA_stats( fd, buf, sizeof(buf), byteSizeCheck, totalBytes, chkSum, statBytesRead );
     DBGPRINT("DBG: calc_sumA returned %d (chkSum=%04u)\n",error,(unsigned)chkSum);
     break;
 case 1:
     snprintf((char*)buf,sizeof(buf)-1,"Bogus file: %s",str);
     perror( (char*)buf );
     break;
 case 3:
     if ( fRepErr ) {
	 fprintf(fRepErr,"Unable to guess end-point: %s\n",str);
     }
     break;
 default:
     break;
 }// end CASE

 bogus = (int)(totalBytes!=byteSizeCheck);
 DBGPRINT_MIN("DBG: totalBytes=%lu/%lu\n",
	      (unsigned long)totalBytes,
	      (unsigned long)byteSizeCheck);

 if ( fRepErr ) {
     if ( error )
	 fprintf(fRepErr,"Erroneous reading%s: %s\n",
		 bogus ? " (I/O error)" : "\0",
		 str);
     else {
	 if ( verboseLevel>=3 ) {
	     fprintf(fRepErr,"%s: read %0.0f%% %u byte(s) / %lu\n",
		     str,
		     totalBytes ? (float)statBytesRead/(float)totalBytes*100.0 : 100.0,
		     statBytesRead,
		     (unsigned long)totalBytes);
	 }
     }
 }
 error = bogus;

 fclose( fIn );
 return error;
}

int do_process_list_sum (gNormList& devnodList, eSumAlgorithm alg, FILE* fOut, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error = 0;
 unsigned i, n = devnodList.N();
 unsigned k;
 t_uint64 uNode;
 bool supportInode( opt.doSupportInode );
 bool doShowFilename( opt.isVerbose==true || opt.doShowAll==true || n>=2 );
 char* str;

#ifdef DEBUG_MIN
 for (i=1; i<=n; i++) {
     printf("Process: '%s' (%d:%lu Hard-linked?%c)\n",
	    devnodList.Str( i ),
	    devnodList.DevNode( i ).GetInt(),
	    (unsigned long)devnodList.DevNode( i ).node,
	    ISyORn(devnodList.DevNode( i ).mask));
 }
#endif //DEBUG_MIN

 for (i=1; i<=n; i++) {
     bool isRepeated( false );
     gDevNode* ptrLazy = &devnodList.DevNode( i );
     int aDev( ptrLazy->GetInt() );
     uNode = ptrLazy->node;
     for (k=(supportInode?1:i); k<i; k++) {
	 gDevNode* ptrComp( &devnodList.DevNode( k ) );
	 if ( aDev==ptrComp->GetInt() && uNode==ptrComp->node ) {
	     isRepeated = true;
	     DBGPRINT("DBG: skipped repeated file: '%s' (uNode=%lu) idx=%u\n",
		      devnodList.Str( i ),
		      (unsigned long)uNode,
		      k);
	     if ( opt.isVeryVerbose && fRepErr )
		 fprintf(fRepErr,"Skipping repeated file: %s\n",devnodList.Str( i ));
	     break;
	 }
     }
     if ( isRepeated ) continue;
     str = devnodList.Str( i );
     t_uint64 totalBytes( 0 );
     t_uint16 chkSum( 0 );

     if ( opt.isVeryVerbose && fRepErr )
	 fprintf(fRepErr,"Processing: %s\n",str);

     switch ( alg ) {
     case e_SumV:
	 error = do_process__sysv_sum_file( str, chkSum, totalBytes );
	 break;
     case e_SumBSD:
	 ASSERTION_FALSE("SumBSD: TODO");
     case e_SumA:
	 error = do_process__suma_file( str, fRepErr, opt.isVeryVerbose ? 3 : 0, chkSum, totalBytes );
	 break;
     default:
	 ASSERTION_FALSE("alg");
     }
     result += error!=0;
     DBGPRINT("DBG: do_process_sum(%s): %d\n",str,error);

     int printMode( (int)doShowFilename + 2*(opt.iValue>0) );
     if ( error ) {
	 if ( opt.isVerbose && fRepErr)
	     fprintf(fRepErr,"Error (code: %d) processing file: %s\n",error,str);
	 if ( opt.isVeryVerbose ) {
	     print_sum_x( fOut, alg, printMode, true, 0, opt.stripLevel, str );
	 }
	 else {
	     if ( opt.doShowAll )
		 print_sum_x( fOut, alg, printMode, true, 0, opt.stripLevel, str );
	 }
	 continue;
     }
     DBGPRINT("DBG: chkSum=%u, totalBytes=%lu\n\n",chkSum,(unsigned long)totalBytes);
     print_sum_x( fOut, alg, printMode, false, chkSum, opt.stripLevel, str );
 }

 return result!=0;
}

int do_build_sum (int level, char* str, gNormList& devnodList, gString& sMsgErr, sOptNorm& opt)
{
 int error = 0;
 t_uint16 nLinks;
 t_inode inode;
 char* strFileName( str );
 char* strErr;

#ifdef gDOS_SPEC
 ASSERTION(strFileName,"strFileName");
 // We avoid having a trailing slash, since Win32 is dummy enough;
 // this will not change anything functionally, but avoids printing, e.g.:
 //	etc\CVS\: No such file or directory
 gString sTempName( strFileName );
 unsigned tempLen( sTempName.Length() );
 if ( sTempName[ tempLen ]==gSLASHCHR )
     sTempName.Delete( tempLen );
 strFileName = sTempName.Str();
#endif //~gDOS_SPEC

 gFileStat probe( strFileName );

 error = probe.HasStat()==false;
 strErr = strerror( probe.lastOpError );
 DBGPRINT("DBG: stat(%s), HasStat?%c, lastOpError=%d, error=%d\n",str,ISyORn(probe.HasStat()),probe.lastOpError,error);
 DBG_SHOW_FSTAT(probe.status);

 if ( error ) {
     sMsgErr.SetEmpty();
     if ( opt.isVeryVerbose ) {
	 sMsgErr.Add( probe.lastOpError );
	 sMsgErr.Add( ": " );
     }
     sMsgErr.Add( strErr );
     return 1;
 }

 if ( probe.IsDirectory() ) {
     sMsgErr.Set( "dir" );
     return -1;
 }
 if ( probe.IsLink() ) {
     sMsgErr.Set( "Link" );
     if ( opt.doShowAll==false ) return -1;
 }
 if ( (probe.status.mode & gFILE_S_IFIFO) ||
      (probe.statusL.mode & gFILE_S_IFIFO) ) {
     sMsgErr.Set( "FIFO" );
     return -1;
 }

 inode = probe.statusL.inode;
 nLinks = probe.status.nLinks;

 gDevNode aDevNode( (t_uint64)inode );
 aDevNode.mask = nLinks>=2;
 devnodList.Add( str, aDevNode );
 DBGPRINT("DBG: added (nLinks=%d): %s\n",nLinks,str);

 return 0;
}
////////////////////////////////////////////////////////////
int do_check (gArg& arg, FILE* fOut, FILE* fRepErr, sOptNorm& opt)
{
 int error;
 int errorSum( 0 ), countChecks( 0 ), countNonStats( 0 );
 unsigned i, n;
 bool doReadTry( opt.doShowAll );
 char* str;
 FILE* fReport( (opt.isVeryVerbose && doReadTry) ? fOut : fRepErr );

 gString sNameRef;
 gNormList lPaths;  // List of files/dirs to check

 ASSERTION(fOut,"fOut");
 ASSERTION((opt.isVerbose && fRepErr) || opt.isVerbose==false,"do_check(1)");

 for (i=1; i<=arg.N(); i++) {
     str = arg.Str(i);

     if ( str==nil || str[0]==0 ) continue;

     gFileStat dummyStat( str );
     gString s( str );
     if ( s[ s.Length() ]!=gSLASHCHR )
	 s.Add( gSLASHCHR );
     str = s.Str();

     if ( dummyStat.status.IsDirectory() ) {
	 t_uint64 byteSize=0;
	 t_uint64 blockSize=0;
	 int recControl=0, nErrors=0;

	 DBGPRINT_MIN("DBG: do_check DIR [%u/%u]: ok?%c: %s\n",
		      i, arg.N(),
		      ISyORn( dummyStat.IsOk() ),
		      str);
	 error = do_du_dir_list( str, fOut, fRepErr, opt, lPaths, byteSize, blockSize, recControl, nErrors );
	 // NEW: recurse checks on dirs
	 if ( error==-1 ) {
	     if ( fRepErr ) {
		 fprintf(fRepErr,"Memory exhausted\n");
	     }
	     return 8;  // Error-code 8: Memory exhausted
	 }
	 errorSum += nErrors!=0;  // New: a directory exists, but if it is not browseable, we set this as an error
     }
     else {
	 lPaths.Add( str );
     }
 }// end FOR (first-cycle): args

 gElem* pCurrent( nil );
 n = lPaths.N();
 if ( n>0 ) {
     lPaths.GetObjectPtr( 1 );
     pCurrent = lPaths.GetCurrentElemPtr();
 }

 for (i=1; i<=n; i++) {
     ASSERTION(pCurrent,"pCurrent");
     str = ((gString*)pCurrent->me)->Str();
     ASSERTION(str,"str");
     pCurrent = pCurrent->next;
     if ( str[0]==0 ) continue;  // Unlikely, due to algorithm used

#ifdef gDOS_SPEC
     // DOS/Win32 is unable to stat "dir\" !
     gString sTemp( str );
     int tempLen( sTemp.Length() );
     if ( sTemp[ tempLen ]==gSLASHCHR ) sTemp[ tempLen ] = 0;
     str = sTemp.Str();
#endif //gDOS_SPEC
     gFileStat dummyStat( str );
     error = gns_errno();
     if ( dummyStat.IsOk()==false ) {
	 if ( opt.isVerbose ) {
	     fprintf(fReport,"Unable to read file/dir (%d): %s\n",
		     error,
		     str);
	 }
	 if ( fReport!=fOut ) {
	     fprintf(fOut,"Unable to read file/dir (%d): %s\n",
		     error,
		     gdr_strip_name_str( str, opt.stripLevel, sNameRef ));
	 }
	 countNonStats++;
	 continue;
     }

     if ( dummyStat.status.IsDirectory() ) {
	 error = 0;  // For now it's sufficient...(maybe cross check dir soft-link?)
     }
     else {
	 if ( doReadTry ) {
	     t_uint32 size( dummyStat.status.size );
	     int bytesToProbe( size > 1024 ? 1024 : (int)size );
	     error = iol_probe_file_read( str, bytesToProbe );
	     countChecks++;
	     DBGPRINT_MIN("DBG: CHECKING #%d: error=%d: %s\n",countChecks,error,str);
	     if ( error ) {
		 errorSum++;
		 if ( opt.isVerbose )
		     fprintf(fReport,"Unable to probe file (%d): %s\n",
			     error,
			     str);
		 if ( fReport!=fOut ) {
		     fprintf(fOut,"Unable to probe file (%d): %s\n",
			     error,
			     gdr_strip_name_str( str, opt.stripLevel, sNameRef ));
		 }
	     }
	     else {
		 if ( opt.isVeryVerbose )
		     fprintf(fReport,"Probed file (%d): %s\n",
			     error,
			     str);
	     }
	 }
	 continue;
     }

     errorSum += error!=0;
     DBGPRINT("DBG: errorSum is currently %d (error=%d): %s\n",
	      errorSum,
	      error,
	      str);
 }//end FOR (i)

 if ( opt.isVerbose ) {
     fprintf(fRepErr,"Checks: %d, failed: %d, non-stats: %d\n",
	     countChecks,
	     errorSum,
	     countNonStats);
 }
 if ( countNonStats>0 ) return 2;
 error = errorSum!=0;
 return error;
}
////////////////////////////////////////////////////////////
int do_du (gArg& arg, FILE* fOut, FILE* fRepErr, sOptNorm& opt)
{
 int error, thisError;
 gList inL;
 gList outL;
 unsigned i, n;
 t_inode inode;
 t_inode inodeCwd;
 char* str;
 char* errorStr;
 gString sCwd;
 gHashTriple inDirHash;
 gKey hKey( gKey::e_UInt64 );

 ASSERTION(fOut,"fOut");

 // Just get the current directory
 gFileControl::Self().GetCwd( sCwd );
 if ( sCwd.IsEmpty() ) {
     fprintf(fRepErr,"Unable to get current directory.\n");
     return 2;
 }
 str = sCwd.Str();
 gXDir aCurDir( str );
 thisError = aCurDir.lastOpError;
 if ( thisError!=0 ) {
     if ( opt.isVerbose ) fprintf(fRepErr,"Unable to open dir: %s, error %d\n",str,thisError);
     return 2;
 }
 inodeCwd = aCurDir.dirStream.dStat.status.inode;

 if ( opt.isChecker ) {
     // The lines within checker file are added to args...for du
     iol_build_list_from_stream( stdin, stderr, arg );
 }

 // Check all locations first
 for (i=1, n=arg.N(), error=0; i<=n; i++) {
     char* strArgOriginal = arg.Str(i);
     str = strArgOriginal;
     if ( str[0]==0 ) continue;
     gDirName myDirName( str );
     str = myDirName.Str();  // Adding a slash
     gXDir aDir( str );
     thisError = aDir.lastOpError;
     bool isNotDirError = thisError==20/*ENOTDIR*/;
     // 'isNotDirError'=True may mean that the user used an argument like:
     //   du /etc/* , but obviously was not planning to check disk-usage of files...
     if ( isNotDirError ) {
	 if ( opt.isVerbose ) fprintf(fRepErr,"Not a directory, skipped: %s\n",strArgOriginal);
	 continue;
     }
     if ( thisError==0 ) {
	 inode = aDir.dirStream.dStat.status.inode;
	 hKey.Set( inode );
	 if ( inDirHash.FindKey( hKey ) ) continue;
	 inDirHash.AddTriple( hKey, str, &aDir );
	 // INFO:
	 //   inodeOfLinkItself = aDir.dirStream.dStat.statusL.inode;
	 ;
	 //if ( inode==inodeCwd )
	 //    inL.Add( "\0" );
	 //else
	 inL.Add( str );
     }
     else {
	 errorStr = gFileControl::Self().ErrorStr( thisError );
	 ASSERTION((opt.isVerbose==false) || (opt.isVerbose && fRepErr),"uops!?");
	 if ( opt.isVeryVerbose ) {
	     fprintf(fRepErr,"Cannot access (%d: %s): %s\n",thisError,errorStr,strArgOriginal);
	 }
	 else {
	     if ( opt.isVerbose ) {
		 fprintf(fRepErr,"Cannot access (%s): %s\n",errorStr,strArgOriginal);
	     }
	 }
	 error++;
     }
 }

 //inDirHash.Show(opt.isVerbose);

 if ( error!=0 ) {
     fprintf(fRepErr,"Number of errors: %d\n",error);
     return 1;
 }

 for (i=1; i<=inL.N(); i++) {
     str = inL.Str(i);
     gFileStat dummyStat( str );
#if gDOS_SPEC
     // Note: Win32 libs do not behave well with DIR\ , i.e. must stat DIR !
     if ( dummyStat.status.mTime==0 ) {
	 gString sWin( str );
	 if ( sWin[ sWin.Length() ]==gSLASHCHR ) {
	     sWin.Delete( sWin.Length() );
	     gFileStat newStat( sWin );
	     dummyStat.status.Copy( newStat.status );
	 }
     }
#endif //gDOS_SPEC
     //if ( str[0]==0 ) str = sCwd.Str();
     t_uint64 byteSize( 0 );
     t_uint64 blockSize( 0 );
     int nErrors( 0 ), recControl( 0 );
     do_du_dir( str, fOut, fRepErr, opt, byteSize, blockSize, recControl, nErrors );
     DBGPRINT_MIN("%8o:%lu\tSUM: %s\n",
		  (unsigned)dummyStat.status.mode,
		  dummyStat.status.mTime,
		  str);
     do_du_print( str, fOut, dummyStat, opt, byteSize, blockSize, false, true );
     // At least one error will result in 'exit(1)'
     if ( nErrors!=0 ) error = 1;
 }

#ifdef DEBUG_HASH
 hKey.Set( (t_uint64)9 );
 opt.myHash.AddTriplePos_dbg( 1, hKey, "test_me" );
#endif //DEBUG_MIN
 // showing hard-links
 if ( opt.DoShowHLinks() ) {
     for (i=0; i<opt.myHash.N(); i++) {
	 gList* pList = opt.myHash.GetHash( i );
	 unsigned n = pList->N();
	 if ( n==0 ) continue;
	 // Whithin this list, group the equal inodes
	 unsigned k, lIdx, lEmptyIdx;
	 t_inode lInode;
	 char* iDescStr;
	 gList* grpList;
	 gKey* grpInodes;
	 grpList = new gList[n];
	 grpInodes = new gKey[n];
	 for (lIdx=0; lIdx<n; lIdx++) grpInodes[lIdx].Set( (t_uint64)MAX_U4B_ULL );
	 for (k=1; k<=n; k++) {
	     gHashElemTriple* aObj;
	     aObj = opt.myHash.GetTriple( i, k );
	     inode = aObj->hKey.GetUInt64();
	     iDescStr = aObj->sDesc.Str();
	     //printf("@: %03u:[%d]:%lx [%s]\n",i,k,(long)inode,iDescStr)
	     // Find inode in any of grpInodes
	     for (lIdx=0, lEmptyIdx=n; lIdx<n; lIdx++) {
		 lInode = grpInodes[lIdx].GetUInt64();
		 if ( grpList[lIdx].N()==0 ) lEmptyIdx = lIdx;
		 if ( inode==lInode ) {
		     grpList[lIdx].Add( iDescStr );
		     break;
		 }
	     }
	     if ( lIdx>=n ) {
		 ASSERTION(lEmptyIdx<n,"lEmptyIdx<n");
		 grpInodes[lEmptyIdx].Set( inode );
		 grpList[lEmptyIdx].Add( iDescStr );
	     }
	 }//end FOR k
	 for (lIdx=0; lIdx<n; lIdx++) {
	     if ( grpList[lIdx].N()==0 ) continue;
	     printf("@@@\t");
	     grpList[lIdx].Show( true );
	 }
	 delete[] grpList;
	 delete[] grpInodes;
     }
 }
 return error;
}
////////////////////////////////////////////////////////////
int do_ddb (gArg& arg, FILE* fOut, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error;
 t_uint16 bufIter;
 t_uint16 bs = opt.bs >= MAX_UINT16_U ? 0 : (t_uint16)opt.bs;
 t_uint32 count = opt.count;
 t_uint32 vRes = 0;
 bool hasInputConstant = false;
 t_uchar chrInputConstant = 0;
 gString sFirst( arg.Str(1) );
 FILE* fIn = stdin;
 FILE* fRepProgress = opt.isVeryVerbose ? stderr : nil;

 ASSERTION(fOut,"fOut");

 //	ddb [input [output]]
 // If only one arg is provided, it is assumed it's the input file-name.
 // If the file-name start by an ash (e.g. '#10') it's assumed the input
 // is a constant ASCII with the decimal value specified (e.g. '#10' is LF, ie Line-feed).
 // Examples:
 //	ddb --bs 1024 --count 2 #0 myout
 //		this will output to file 'myout' 1024 octets ASCII 0 twice, i.e. 2 Kbytes.
 ;

 gUCharBuffer sBuf( bs );
 // buffer created here, because if it is non-asserted: no output is created for quitting...

 hasInputConstant = sFirst[1]=='#';
 if ( hasInputConstant ) {
     fIn = nil;
     if ( gStrControl::Self().ConvertToUInt32( sFirst.Str()+1, vRes ) ||
	  vRes>255 ) {
	 if ( fRepErr )
	     fprintf(fRepErr,"Invalid ASCII: %lu\n",(unsigned long)vRes);
	 return 1;
     }
     chrInputConstant = (t_uchar)vRes;
     for (bufIter=0; bufIter<sBuf.size; bufIter++) {
	 DBGPRINT_MIN("DBG: initzing buf(%u/%u, size=%u): ASCII#%u\n",
		      bufIter,
		      bs,
		      sBuf.size,
		      (unsigned)chrInputConstant);
	 sBuf.uBuf[bufIter] = chrInputConstant;
     }
 }
 else {
     fIn = stdin;
     if ( sFirst.IsEmpty()==false ) {
	 fIn = fopen( sFirst.Str(), "rb" );
	 if ( fIn==nil ) {
	     if ( fRepErr )
		 fprintf(fRepErr,"Unable to read file: %s\n",sFirst.Str());
	     return 2;
	 }
	 DBGPRINT_MIN("DBG: fopen for read: %s\n",sFirst.Str());
     }
     DBGPRINT_MIN("DBG: buf(bs=%u, size=%u)\n",
		  bs,
		  sBuf.size);
 }

 error = do_dd_io( fIn, fOut, fRepErr, fRepProgress, sBuf, count );
 DBGPRINT_MIN("DBG: do_dd_io returned %d\n",error);
 if ( fIn!=nil && fIn!=stdin ) fclose( fIn );
 result = error;
 DBGPRINT("DBG: do_ddb with error:%d, returns %d\n",error,error!=0);
 return error!=0;
}

int do_prepare_ddb (gArg& arg, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error;
 unsigned n = arg.N();
 gString sSecond( arg.Str(2) );
 char* outFilename( sSecond.Str() );
 FILE* fNewOut( stdout );

 if ( n>=2 ) {
     fNewOut = fopen( outFilename, "wb" );
     if ( fNewOut==nil ) {
	 if ( fRepErr )
	     fprintf(fRepErr,"Unable to write file: %s\n",outFilename);
	 return 2;
     }
     DBGPRINT_MIN("DBG: fopen for write: %s\n",outFilename);
     if ( n>2 ) {
	 if ( fRepErr )
	     fprintf(fRepErr,"Ignored %u arguments (%s ...)\n",
		     n-2,
		     arg.Str(3));
     }
 }

 error = do_ddb( arg, fNewOut, fRepErr, opt, result );
 if ( error && fRepErr ) {
     fprintf(fRepErr,"Unable to create file properly (%d): %s\n",error,outFilename);
 }

 // Fail-safe: output is removed if there was an error
 if ( fNewOut!=nil && fNewOut!=stdout ) {
     fclose( fNewOut );
     if ( outFilename[0]!=0 && error>0 ) remove( outFilename );
 }

 return error;
}
////////////////////////////////////////////////////////////
int do_stat (gArg& arg, FILE* fOut, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error = 0;
 unsigned i, n=arg.N();
 char* str;

 for (i=1; i<=n; i++) {
     str = arg.Str( i );
     if ( do_stat_print( str, fOut, fRepErr, opt, result ) )
	 error++;
 }
 return error;
}
////////////////////////////////////////////////////////////
int do_sumv (gArg& arg, FILE* fOut, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error = 0;
 int countErrors = 0;
 unsigned i, n=arg.N();
 char* str;
 char* strErr;
 gString sMsgErr;

 gNormList devnodList;

 ASSERTION(fOut,"fOut");

 for (i=1; i<=n; i++) {
     str = arg.Str( i );
     error = do_build_sum( 0, str, devnodList, sMsgErr, opt );
     if ( error==-1 ) {
	 DBGPRINT("DBG: skipping dir/%s: '%s'\n",sMsgErr.Str(),str);
	 continue;
     }
     if ( error ) {
	 strErr = sMsgErr.Str();
	 DBGPRINT("DBG: strErr='%s'\n",strErr);
	 if ( fRepErr ) fprintf(fRepErr,"%s: %s\n",str,strErr);
	 countErrors++;
     }
 }

 if ( n==0 ) {
     t_uint64 totalBytes( 0 );
     t_uint16 chkSum( 0 );
     error = do_process__sysv_sum_file( nil, chkSum, totalBytes );
     if ( fOut )
	 fprintf(fOut,"%u\n",chkSum);
 }
 else {
     error = do_process_list_sum( devnodList, e_SumV, fOut, fRepErr, opt, result );
 }
 DBGPRINT("DBG: do_process_sum [sumV] returned %d\n",error);

 return error;
}
////////////////////////////////////////////////////////////
int do_suma (gArg& arg, FILE* fOut, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error = 0;
 int countErrors = 0;
 unsigned i, n=arg.N();
 char* str;
 char* strErr;
 gString sMsgErr;

 gNormList devnodList;

 ASSERTION(fOut,"fOut");

 if ( n==0 && iol_build_list_from_stream( stdin, stderr, arg )>0 )
     n = arg.N();

 for (i=1; i<=n; i++) {
     str = arg.Str( i );
     error = do_build_sum( 0, str, devnodList, sMsgErr, opt );
     if ( error==-1 ) {
	 DBGPRINT("DBG: skipping dir/%s: '%s'\n",sMsgErr.Str(),str);
	 if ( opt.isVerbose && fRepErr )
	     fprintf(fRepErr,"Skipping %s: '%s'\n",sMsgErr.Str(),str);
	 continue;
     }
     if ( error ) {
	 strErr = sMsgErr.Str();
	 DBGPRINT("DBG: strErr='%s'\n",strErr);
	 if ( fRepErr ) fprintf(fRepErr,"%s: %s\n",str,strErr);
	 countErrors++;
     }
 }

 error = do_process_list_sum( devnodList, e_SumA, fOut, fRepErr, opt, result );
 DBGPRINT("DBG: do_process_sum [sum-A] returned %d\n",error);

 return error;
}

int do_chk_suma (gArg& arg, FILE* fOut, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error( 0 );
 int countErrors( 0 );
 unsigned i, n( 0 );
 char* str;
 gString sFirst( opt.sCheckerFile );
 gList inList;
 gList chkLines;
 gElem* pCurrent;

 gNormList devnodList;

 ASSERTION(fOut,"fOut");
 ASSERTION(arg.N()==0,"arg.N()==0");  // Just to re-enforce this arg is not needed

 if ( sFirst.IsEmpty() ) {
     iol_build_list_from_stream( stdin, stderr, inList );
 }
 else {
     FILE* fIn( fopen( sFirst.Str(), "rt" ) );
     if ( fIn==nil ) return 2;
     iol_build_list_from_stream( fIn, stderr, inList );
     fclose( fIn );
 }

 n = inList.N();
 DBGPRINT("DBG: do_chk_suma with %u lines to check\n",n);

 for (i=1; i<=n; i++) {
     // Check consistency of all check lines:
     gString s( inList.Str(i) );
     error = sul_add_checkline( s, chkLines );
     DBGPRINT("DBG: sul_add_checkline(%s) returned %d\n",
	      s.Str(),
	      error);
     if ( error>0 ) countErrors++;
 }

 if ( countErrors ) {
     if ( fRepErr ) {
	 fprintf(fRepErr,"Got %u check-line(s), but with %d error(s)\n",
		 n,
		 countErrors);
     }
     return 1;
 }

 n = chkLines.N();
 str = chkLines.Str( 1 );
 pCurrent = chkLines.GetElementPtr( 0 );

 for (i=1; i<=n; i++, pCurrent=pCurrent->next) {
     t_uint64 totalBytes( 0 );
     t_uint16 chkSum( 0 );
     gSumLine* sumLine( (gSumLine*)pCurrent->me );

     str = sumLine->Str();  // str = chkLines.Str(i) => would do as well, but slower!
     error = do_process__suma_file( str, fRepErr, opt.isVeryVerbose ? 3 : 0, chkSum, totalBytes );
     if ( error==0 ) {
	 error = (sumLine->uCheck!=chkSum)*8;
	 DBGPRINT("DBG: %u/%u\t%s\n",
		  sumLine->uCheck, chkSum,
		  str);
     }
     DBGPRINT("DBG: do_process__suma_file(%s) returned %d\n",str,error);
     countErrors += error!=0;
     if ( opt.isVerbose ) {
	     fprintf(fOut,"%s\t%s\n",
		     error ? "ERROR" : "OK",
		     str);
	     if ( opt.isVeryVerbose && fOut!=stdout )
		 fprintf(stdout,"%s\t%s\n",
			 error ? "ERROR" : "OK",
			 str);
     }
 }// end FOR (each check-line)

 if ( opt.isVeryVerbose ) {
     ASSERTION(fRepErr,"fRepErr");
     fprintf(fRepErr,"%u file(s), %d error(s)\n",n,countErrors);
 }

 return countErrors!=0;
}
////////////////////////////////////////////////////////////
int do_pwd (gArg& arg, FILE* fOut, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error( 0 );
 unsigned i( 0 ), n( arg.N() );
 gUCharBuffer bBuf( 4096 );

 ASSERTION(fOut,"fOut");

 char* strPwd = gio_getcwd( (char*)bBuf.uBuf, bBuf.size );
 error = gStorageControl::Self().lastOpError;
 DBGPRINT("DBG: strPwd='%s' (max:%u, error=%d)\n",strPwd,bBuf.size,error);
 if ( error ) return error;

 for ( ; i<=n; ) {
     gString sName( strPwd );
     if ( gdr_strip_name( opt.stripLevel, 1, '\0', sName )<0 ) {
	 if ( fRepErr )
	     fprintf(fRepErr,"Fallback stripped name: %s\n",sName.Str());
     }
     fprintf(fOut,"%s\n",sName.Str());
     i++;
     strPwd = arg.Str( i );
 }

 return 0;
}
////////////////////////////////////////////////////////////
int do_touchb (gArg& arg, FILE* fOut, FILE* fRepErr, sOptNorm& opt, int& result)
{
 int error( 0 );
 int errorCode( -1 );
 unsigned i( 1 ), n( arg.N() );

 if ( n<=0 || n>3 ) {
     if ( fRepErr ) fprintf(fRepErr,"Invalid #params: %u\n",n);
     return 1;
 }

 char* strFilename( arg.Str( i++ ) );
 t_stamp currentTime( gio_current_time() );
 t_uint32 v1( (t_uint32)currentTime ), v2( v1 );
 if ( n>=2 ) {
     if ( gStrControl::Self().ConvertToUInt32( arg.Str( i++ ), v1 ) ) {
	 if ( fRepErr ) fprintf(fRepErr,"Invalid modification time.\n");
	 return 1;
     }
     if ( n>=3 ) {
	 if ( gStrControl::Self().ConvertToUInt32( arg.Str( i++ ), v2 ) ) {
	     if ( fRepErr ) fprintf(fRepErr,"Invalid access time.\n");
	     return 1;
	 }
     }
 }
 else {
     if ( fOut ) {
	 fprintf(fOut,"Considering current time: %lu\n",(unsigned long)v1);
     }
 }

 DBGPRINT_MIN("DBG: do_touchb: n=%u: %s, %lu, %lu; currentTime: %lu\n",
	      n,
	      strFilename,
	      (unsigned long)v1,
	      (unsigned long)v2,
	      (unsigned long)currentTime);
 error = gio_file_touch_check( strFilename, (t_stamp)v2, (t_stamp)v1, errorCode );
 DBGPRINT("DBG: do_touchb returns %d, errorCode=%d\n",
	  error,
	  errorCode);
 return error;
}
////////////////////////////////////////////////////////////
int do_sleep (gArg& arg, FILE* fOut, FILE* fRepErr, int showMode, sOptNorm& opt, int& result)
{
 unsigned iter;
 int error( 0 );
 int tic( 0 );
 t_uint32 aSec( 0 );
 t_uint32 aMilli( 0 );

 if ( gStrControl::Self().ConvertToUInt32( arg.Str(1), aSec ) ) {
     if ( fRepErr ) {
	 fprintf(fRepErr,"Wrong seconds specified.\n");
	 return 1;
     }
 }
 if ( gStrControl::Self().ConvertToUInt32( arg.Str(2), aMilli ) ) {
     if ( fRepErr ) {
	 fprintf(fRepErr,"Wrong milliseconds specified.\n");
	 return 1;
     }
 }
 arg.Delete( 1, 2 );
 if ( arg.N() ) {
     for (iter=1; iter<=arg.N(); iter++) {
	 printf("%s%s",
		iter>1 ? " " : "\0",
		arg.Str(iter));
     }
     printf("\n");
 }

 switch ( showMode ) {
 case 0:
     error = gFileControl::Self().SecSleep( aSec );
     break;

 case 1:
 default:
     for (tic=(int)aSec; tic>=0; tic--) {
	 error = gFileControl::Self().SecSleep( 1 );
	 if ( fOut ) {
	     fprintf(fOut,".");
	     fflush( fOut );
	 }
     }
     if ( fOut ) fprintf(fOut,"\n");
     break;
 }

 // nanosleep sleeps for 10^-9 seconds (10^9/10^3=10^6):
 if ( aMilli>1000 ) aMilli = 1000;
 aMilli *= 1000;
#ifdef gDOS_SPEC
 for (iter=0; iter<=aMilli; iter++) {
     gFileControl::Self().SecSleep( 0 );  // Rough milliseconds on Win32!
 }
#else
 struct timespec req;
 aMilli *= 1000;
 req.tv_sec = 0;
 req.tv_nsec = (long)aMilli;
 nanosleep( &req, nil );
#endif //gDOS_SPEC

 return error;
}
////////////////////////////////////////////////////////////
int go (char** argv, char** envp, int& result)
{
 int error;
 gArg arg( argv, envp );
 sOptNorm opt;
 char* str;
 long value;

 unsigned n;
 FILE* fOut( stdout );
 FILE* fRepErr( stderr );

 result = 0;

 arg.AddParams( 1, "-vv|--verbose -h|--help --version\
 -a|--all\
 -c:%s|--checker:%s\
 -o:%s|--output:%s\
 -z:%d|--value:%d\
 -b|--bytes -k|--kilobytes -m|--megabytes --summarize --exact --hard-links\
 --no-sort\
 --max-depth=%u\
 -s=%d|--strip=%d\
 --bs:%u --count:%u" );

 // Parse command line
 error = arg.FlushParams();
 n = arg.N();
 arg.FindOptionOccurr( "verbose", opt.isVerbose, opt.isVeryVerbose );

 gString firstCmd( arg.Str(1) );
 char* cmdStr = firstCmd.Str();
 arg.Delete( 1, 1 );

 if ( arg.FindOption('h') || error!=0 ) {
     print_help( str = arg.Program() );
     if ( error==0 && n==1 && cmdStr[0]!='-' ) {
	 print_help_command( str, cmdStr );
     }
     else {
	 printf("For command help, use: %s command --help\n",str);
     }
     return 0;
 }
 if ( arg.FindOption("version") ) {
     print_version( arg.Program() );
     return 0;
 }

 opt.doShowAll = arg.FindOption('a');

 opt.duOpt.sortCode = (int)(arg.FindOption("no-sort")==false);
 opt.duOpt.doByByteSize = arg.FindOption('b');
 opt.duOpt.doByMegaBytes = arg.FindOption('m');
 opt.duOpt.doByBlockSize = arg.FindOption('k') || opt.duOpt.doByMegaBytes;
 opt.duOpt.doSummarize = arg.FindOption("summarize");
 opt.duOpt.doExact = arg.FindOption("exact");
 opt.duOpt.doCountHardLinks = arg.FindOption("hard-links");

 arg.FindOption( "max-depth", opt.duOpt.aMaxDepth );

 if ( arg.FindOption( "value", opt.iValue ) ) {
     opt.duOpt.zValue = opt.iValue;
 }

 if ( arg.FindOption( "strip", opt.stripLevel ) ) {
     if ( opt.stripLevel<-1 ) {
	 fprintf(fRepErr,"Invalid strip level: %d (minimum is -1)\n",opt.stripLevel);
	 return 1;
     }
 }

 if ( arg.FindOption( "output", opt.sOutput ) ) {
     fOut = fopen( opt.sOutput.Str(), "wb" );
     if ( fOut==nil ) {
	 fprintf(fRepErr,"Cannot write file: %s\n",opt.sOutput.Str());
	 return 1;
     }
 }

 if ( arg.FindOption( "checker", opt.sCheckerFile ) ) {
     FILE* fCheck( fopen( opt.sCheckerFile.Str(), "rb" ) );
     bool tryStdin( opt.sCheckerFile.Match("-") || opt.sCheckerFile.IsEmpty() );

     opt.isChecker = true;
     if ( tryStdin ) {
	 opt.sCheckerFile.SetEmpty();
     }
     else {
	 if ( fCheck ) {
	     fclose( fCheck );
	 }
	 else {
	     fprintf(fRepErr,"Cannot use checker file: %s\n",opt.sCheckerFile.Str());
	     return 2;
	 }
     }
 }

 if ( arg.FindOption( "bs", value ) ) {
     if ( value<=0 || value>=(long)MAX_UINT16_U ) {
	 // Previous max: MAX_DLINT32
	 fprintf(fRepErr,"Invalid block-size specified: %ld, max is %u\n",value,MAX_UINT16_U-1);
	 return 0;
     }
     opt.bs = (t_uint32)value;
 }
 if ( arg.FindOption( "count", value ) ) {
     if ( value<=0 || value>=MAX_DLINT32 ) {
	 fprintf(fRepErr,"Invalid count specified: %ld\n",value);
	 return 0;
     }
     opt.count = (t_uint32)value;
 }

 opt.duOpt.viewMask = opt.isVerbose + 2*opt.isVeryVerbose;

 if ( n==0 ) {
     fprintf(fRepErr,"Empty command. Use --help to display usage.\n");
     return 0;
  }

 if ( opt.doShowAll ) {
     opt.duOpt.doShowAll = true;
     if ( opt.duOpt.doSummarize ) {
	 fprintf(fRepErr,"%s: cannot both summarize and show all entries.\n",cmdStr);
	 return 1;
     }
 }

 if ( strcmp(cmdStr,"check")==0 ) {
     opt.isChecker = true;
     error = do_check( arg, fOut, fRepErr, opt );
     return error;
 }
 if ( strcmp(cmdStr,"du")==0 ) {
     error = do_du( arg, fOut, fRepErr, opt );
     return error;
 }
 if ( strcmp(cmdStr,"ddb")==0 ) {
     error = do_prepare_ddb( arg, opt.isVerbose ? fRepErr : nil, opt, result );
     return error;
 }
 if ( strcmp(cmdStr,"stat")==0 ) {
     error = do_stat( arg, fOut, opt.isVerbose ? fRepErr : nil, opt, result );
     return error;
 }
 if ( strcmp(cmdStr,"sumv")==0 ) {
     error = do_sumv( arg, fOut, opt.isVerbose ? fRepErr : nil, opt, result );
     return error;
 }
 if ( strcmp(cmdStr,"suma")==0 ) {
     if ( opt.isChecker && arg.N() ) {
	 fprintf(fRepErr,"suma check has no args.\n");
	 return 1;
     }
     if ( opt.isChecker )
	 error = do_chk_suma( arg, fOut, opt.isVerbose ? fRepErr : nil, opt, result );
     else
	 error = do_suma( arg, fOut, opt.isVerbose ? fRepErr : nil, opt, result );
     return error;
 }
 if ( strcmp(cmdStr,"pwd")==0 ) {
     error = do_pwd( arg, fOut, opt.isVerbose ? fRepErr : nil, opt, result );
     return error;
 }
 if ( strcmp(cmdStr,"touchb")==0 ) {
     error = do_touchb( arg, fOut, opt.isVerbose ? fRepErr : nil, opt, result );
     return error;
 }
 if ( strcmp(cmdStr,"sleep")==0 ) {
     error = do_sleep( arg, stdout, opt.isVerbose ? fRepErr : nil, opt.iValue!=0, opt, result );
     return error;
 }

 fprintf(fRepErr,"Invalid command: %s\n",cmdStr);
 return result = 1;
}
////////////////////////////////////////////////////////////
int main (int argc, char** argv, char** envp)
{
 int error, result=0;

 gINIT;
 error = go( argv, envp, result );
 gEND;

 DBGPRINT_MIN("go: %d, result=%d\n",error,result);

 return error;
}

