// gxdir.cpp -- Version 1.0
//
// Based on gdir.cpp/h of gobj, but allows long-sized files.

#include <string.h>
#include <errno.h>

#include "gXDir.h"
////////////////////////////////////////////////////////////
// Static members
// ---------------------------------------------------------
// <>
////////////////////////////////////////////////////////////
gSeq::gSeq ()
{
}

gSeq::~gSeq ()
{
}

int gSeq::InsertObject (gStorage* ptrNew)
{
 if ( ptrNew==nil ) return -1;

 if ( size==0 ) {
     AppendObject( ptrNew );
     return 1;
 }

 gElem* keepStart( pStart );
 gElem* newElem( new gElem( ptrNew ) );
 if ( newElem==nil ) return -1;

 pStart = newElem;
 newElem->next = keepStart;
 keepStart->prev = newElem;
 size++;

 pCurrent = pStart;
 return 2;
}

int gSeq::SetSize (unsigned aSize)
{
 size = aSize;
 return 0;
}

int gSeq::SetStartPtr (gElem* ptrElem)
{
 pStart = ptrElem;
 return pStart==nil;
}

int gSeq::SetEndPtr (gElem* ptrElem)
{
 pEnd = ptrElem;
 return pEnd==nil;
}
////////////////////////////////////////////////////////////
gXFilePtr::gXFilePtr (gFileSysName* aPtrSys)
    : ptrSys( aPtrSys )
{
 kind = e_StoreExtend;
}

gXFilePtr::~gXFilePtr ()
{
 // Note: ptrSys is NEVER deleted!
}

char* gXFilePtr::Str ()
{
 if ( ptrSys==nil ) return nil;
 return ptrSys->Str();
}

gFileSysName* gXFilePtr::SysName ()
{
 ASSERTION(ptrSys,"ptrSys");
 return ptrSys;
}

int gXFilePtr::AddToList (gSeq& list)
{
 ASSERTION_FALSE("TODO: CHECK");
 gXFilePtr* ptrNew( new gXFilePtr( ptrSys ) );
 if ( ptrNew==nil ) return -1;
 list.AppendObject( ptrNew );
 return 0;
}

int gXFilePtr::AddSysElements (gElem* ptrElem, gSeq& list)
{
 gXFilePtr* ptrNew;
 for ( ; ptrElem; ptrElem=ptrElem->next) {
     ptrNew = new gXFilePtr( (gFileSysName*)ptrElem->me );
     if ( ptrNew==nil ) return -1;
     list.AppendObject( ptrNew );
 }
 return 0;
}
////////////////////////////////////////////////////////////
gXDir::gXDir (char* dName, int mask)
    : gDir( dName, false ),
      ioFails( -1 ),
      maskValue( mask ),
      sorted( nil )
{
 DBGPRINT_MIN("DBG: OPENED?%c gXDir(%s):lastOpError=%d\n",
	      ISyORn( dirStream.IsOpened() ),
	      dirStream.Name(),
	      lastOpError);
 if ( dName!=nil ) {
     thisDelete();
     thisScanDirXted( mask );
 }

 //printf("gXDir at '%s' is: #%u: ",dName,N()); Show();
 //printf("errUnstatL: "); errUnstatL.Show();
 ioFails = (int)errUnstatL.N();
}

gXDir::~gXDir ()
{
 DBGPRINT_MIN("DBG: delete sorted: %d\n",
	      sorted ? (int)sorted->N() : -1);
 delete sorted;
}

gElem* gXDir::StartSortedPtr ()
{
 if ( sorted ) {
     return sorted->StartPtr();
 }
 return StartPtr();
}

int gXDir::InsertOrderedSys (gFileSysName* fSysPtr, gSeq& list)
{
 gElem* ptrElem( list.StartPtr() );
 gElem* ptrFirst( ptrElem );
 gElem* ptrNext( ptrElem );
 gElem* newElem( nil );
 gStorage* myObj( nil );
 char* givenName( fSysPtr->Str() );
 gXFilePtr* ptrNew;

 if ( fSysPtr==nil ) return -1;

 ptrNew = new gXFilePtr( fSysPtr );
 if ( ptrNew==nil ) return -2;

 if ( ptrElem==nil ) {
     list.AppendObject( ptrNew );
     return 1;
 }

 ptrNext = ptrElem->next;
 if ( ptrNext==nil ) {
     myObj = ptrElem->me;
     if ( gxdir_string_compare( givenName, myObj->Str() )<0 ) {
	 list.InsertObject( ptrNew );
	 return 2;
     }
 }

 for ( ; ; ) {
     myObj = ptrElem->me;
     DBGPRINT_MIN("\t%d\t#%u: { %s, %s }\n",strcmp( givenName, myObj->Str() ),list.N(),myObj->Str(),givenName);

     if ( gxdir_string_compare( givenName, myObj->Str() )<0 ) {
	 if ( ptrElem==ptrFirst ) {
	     list.InsertObject( ptrNew );
	     return 3;
	 }

	 newElem = new gElem( ptrNew );
	 ASSERTION(newElem,"newElem");

	 pCurrent = ptrElem->prev;
	 ptrElem->prev = newElem;
	 newElem->next = ptrElem;
	 ptrElem = pCurrent;
	 if ( ptrElem ) {
	     ptrElem->next = newElem;
	 }
	 else {
	     ASSERTION_FALSE("uops: insert");
	 }
	 newElem->prev = ptrElem;

	 list.SetSize( list.N()+1 );
	 return 4;
     }

     ptrElem = ptrNext;
     if ( ptrElem==nil ) break;
     ptrNext = ptrElem->next;
 }

 // It's the last
 list.AppendObject( ptrNew );
 return 5;
}

gSeq* gXDir::thisSortDir (int sortCode)
{
 int code;
 unsigned k, n( N() );
 unsigned nDirs( 0 ), ofIdx, ofN;
 char* aName;
 gSeq* order;
 gElem* ptrElem;
 gFileSysName* fSysPtr( nil );
 //gFileStat* pStat( nil );
 gXFilePtr* ptrNew;

 short dirCode, areDirs[ n+1 ];

 order = new gSeq;
 if ( order==nil ) return nil;

 if ( sortCode<=0 ) {
     gXFilePtr copy;
     copy.AddSysElements( StartPtr(), *order );
     return order;
 }

 /*
 printf("DBG: Before-sort >>>\n");
 for (k=1, ptrElem=StartPtr(); k<=n; k++, ptrElem=ptrElem->next) {
     fSysPtr = (gFileSysName*)ptrElem->me;
     aName = fSysPtr->Str();
     printf("DBG: %u/%u: dir?%c: 0x%p: %s\n",
	    k, n,
	    ISyORn( fSysPtr->IsDirectory() ),
	    fSysPtr,
	    aName);
 }
 printf("DBG: End <<<\n");
 */

 // Sort directories first

 // ...so start indexing:
 for (k=1, ptrElem=StartPtr(); k<=n; k++, ptrElem=ptrElem->next) {
     fSysPtr = (gFileSysName*)ptrElem->me;
     //pStat = fSysPtr->GetStatPtr();
     aName = fSysPtr->Str();
     if ( (dirCode = (short)fSysPtr->IsDirectory())==1 ) {
	 code = InsertOrderedSys( fSysPtr, *order );
	 nDirs++;
	 DBGPRINT_MIN("DBG: InsertOrderedSys, code=%d: '%s'\n",code,aName);
	 // { gElem* ptrElem; printf("DBG: list: #%u: ",order->N()); for (ptrElem=order->StartPtr(); ptrElem; ptrElem=ptrElem->next) printf("%s ",ptrElem->me->Str()); printf("!\n\n"); }
     }
     areDirs[ k ] =  dirCode;
 }

 gSeq ordFiles;

 for (k=1, ptrElem=StartPtr(); k<=n; k++, ptrElem=ptrElem->next) {
     if ( areDirs[ k ] ) continue;
     fSysPtr = (gFileSysName*)ptrElem->me;
     InsertOrderedSys( fSysPtr, ordFiles );
 }

 // Not so optimized for files, but here goes along...easily:
 for (ofIdx=1, ofN=ordFiles.N(), ptrElem=ordFiles.StartPtr(); ptrElem; ptrElem=ptrElem->next) {
     ptrNew = (gXFilePtr*)ptrElem->me;
     fSysPtr = ptrNew->ptrSys;
     ptrNew = new gXFilePtr( fSysPtr );
     ASSERTION(ptrNew,"ptrNew");
     order->AppendObject( ptrNew );
 }

 /*
 printf("DBG: Now (%u=%u) >>>\n",n,order->N());
 for (k=1, n=order->N(), ptrElem=order->StartPtr(); k<=n; k++, ptrElem=ptrElem->next) {
     ASSERTION(ptrElem && ptrElem->me,"now: ptrElem");
     ptrNew = (gXFilePtr*)ptrElem->me;
     fSysPtr = ptrNew->ptrSys;
     aName = fSysPtr->Str();
     printf("DBG: %u/%u: dir?%c: 0x%p: %s\n",
	    k, n,
	    ISyORn( fSysPtr->IsDirectory() ),
	    fSysPtr,
	    aName);
 }
 printf("DBG: End <<<\n");
 */

 return order;
}

int gXDir::thisScanDirXted (int mask)
{
 bool isOk, isDir;
 struct dirent tDirEnt;
 struct dirent* dp;
 char* s( nil );
 int counter=0;
 DIR* apDir;

 apDir = dirStream.pdir;
 ASSERTION(dirStream.dirName.IsEmpty()==false,"thisScanDir(1)");

 if ( apDir==nil ) {
     lastOpError = errno;
     return lastOpError;
 }

 for (lastOpError=0; ; counter++) {
     gFileStat aStat;
     dp = readdir( apDir );
     if ( counter==0 && dp==NULL ) lastOpError = errno;
     if ( dp==NULL ) break;
     memset( &tDirEnt, 0, sizeof(tDirEnt) );
     memcpy( &tDirEnt, dp, sizeof(tDirEnt) );
     // d_reclen is the size, d_name the name itself
     s = (char*)tDirEnt.d_name;
     if ( strcmp(s,".")==0 || strcmp(s,"..")==0 ) continue;
     gString sTemp( sDirName.Name() );
     sTemp.Add( gSLASHCHR );
     sTemp.Add( s );
     aStat.Update( sTemp.Str() );
     isOk = aStat.HasStat();
     DBGPRINT_MIN("DBG: gdir: '%s', isOk?%c: %s\n",
		  s,
		  ISyORn( isOk ),
		  sTemp.Str());
     if ( isOk==false ) {
	 errUnstatL.Add( sTemp );
	 continue;
     }
     isDir = aStat.statusL.IsDirectory();
     thisAddStatName( isDir?e_FSysDir:e_FSysFile, s, aStat );
 }
 DBGPRINT("gXDir: counter=%d: '%s' : lastOpError=%d, last='%s'\n",
	  counter,
	  dirStream.Name(),
	  lastOpError,
	  s);
 if ( s==nil ) {
     if ( lastOpError==0 ) lastOpError = 14;
 }
 return lastOpError;
}
////////////////////////////////////////////////////////////
char* gxdir_substed_strs (char* s)
{
 char chr;
 char* result( s );
 int idx( 0 );
 for ( ; (chr = *s)!=0; s++) {
     // Replace '\' (backslash) by ASCII 31d
     result[ idx++ ] = (chr=='/' || chr=='\\') ? 31 : chr;
 }
 return result;
}

int gxdir_string_compare (char* s1, char* s2)
{
 if ( s1==nil || s2==nil ) return -1;

 gString aString1( s1 );
 gString aString2( s2 );
 
 if ( gxdirSort ) {
     return strcmp( gxdir_substed_strs( aString1.Str() ), gxdir_substed_strs( aString2.Str() ) );
 }
 return -1;
}
////////////////////////////////////////////////////////////

