// gStrings.cpp -- Version 1.0

#define thisProgramVersion "Version 1.0"
#define thisProgramCopyright "Prized Season & Sons"
#define thisProgramYear 2009
#define thisProgramCopy "This is free software (GPL)\n\
There is no warranty, not even for MERCHANTABILITY or\n\
FITNESS FOR A PARTICULAR PURPOSE."

#include <stdio.h>
#include <string.h>
#include "garg.h"
#include "gfile.h"

#include "gSupStrUni.h"

#ifdef gDOS_SPEC
#include <windows.h>
#endif //gDOS_SPEC

// Global (uggly, not using libgobjeio...)
gOptUnCode* ptrUncode;
sUnCodeBin* ptrUcBin;
////////////////////////////////////////////////////////////
struct sOptStrgs {
    sOptStrgs ()
	: isVerbose( false ),
	  isVeryVerbose( false ),
	  isOutDOS( false ),
	  isDumpLinear( false ),
	  doSkip8Bit( false ),
	  doExtended( false ),
	  doViewAll( false ),
	  caseKind( 0 ) {
    }
    ~sOptStrgs () {
    }

    bool isVerbose, isVeryVerbose;
    bool isOutDOS;
    bool isDumpLinear;
    bool doSkip8Bit;
    bool doExtended;
    bool doViewAll;

    int caseKind;	// 0:none; 1:up; 2:down/lower-case

    gString sDumpFile;
    gString sCase;
};
////////////////////////////////////////////////////////////
int print_help (char* progStr)
{
 char
     *msgHelp = "%s - %s\n\n\
Usage:\n\
        %s command [options] [file ...]\n\
\n\
Options are:\n\
   -h              This help (or --help / --version)\n\
   -v              Verbose (use twice, more verbose)\n\
   -8              Skip 8bit ISO 8859-1 latin chars (or --skip-8bit)\n\
   -c S            Case S (or --case S)\n\
   -o X            Output to file X (or --output)\n\
   --dos-mode      DOS-mode (CR&LF)\n\
   --dump-linear   Linear dump\n\
   --view          View printable chars\n\
\n\
gstrings uses ISO 8859-1 default as per UC 0x0...0xFF:\n\
	%s\n\
\n\
Cases are:\n\
	no case (no option -c)\n\
	normal: convert to best 7bit char, if possible\n\
	up, or upper: convert to uppercase\n\
	down, or lower: convert to lowercase\n\
\n\
Examples:\n\
	gstrings -c normal mytext.txt\n\
		shows the text file in best-approximate 7bit chars.\n\
	gstrings -c up mytext.txt\n\
		shows the text file in upper-case\n\
\n";

 // '-e' is not used yet:::
 //  -e              Extended: converts chars into 8bit

 printf(msgHelp,
	progStr,
	thisProgramVersion,
	progStr,
	ptrUncode->UniThbFileStr());
 return 0;
}
////////////////////////////////////////////////////////////
int print_version (char* progStr)
{
 char
     *msgVersion = "%s - %s\n\
Written by Henrique Moreira.\n\
\n\
Copyright (C) %u %s.\n\
%s\n";

 printf(msgVersion,
	progStr,
	thisProgramVersion,
	thisProgramYear,thisProgramCopyright,
	thisProgramCopy);
 return 0;
}
////////////////////////////////////////////////////////////
// glib required functions
////////////////////////////////////////////////////////////
int gglobal_EventTerminate (int v)
{
 return 0;
}
////////////////////////////////////////////////////////////
// Aux functions
////////////////////////////////////////////////////////////
char* strg_nl (bool showNL, sOptStrgs& opt)
{
 static char shortBuf[ 8 ];
 static int iter;

 shortBuf[ iter=0 ] = 0;
 if ( showNL ) {
     if ( opt.isOutDOS ) {
	 shortBuf[ iter++ ] = '\r';
     }
     shortBuf[ iter++ ] = '\n';
     shortBuf[ iter ] = 0;
 }
 return shortBuf;
}
////////////////////////////////////////////////////////////
// Specific functions
////////////////////////////////////////////////////////////
int go_file (FILE* fIn, FILE* fOut, FILE* fRepErr,
	     sUnCodeBin& uc, gOptUnCode& unCode, sOptStrgs& opt)
{
 int error;
 unsigned maxChr( uc.n ), unChr;
 t_uchar chr( '\0' );
 t_uchar refChrUse( '\0' );
 t_uchar lastChr( '\0' );
 bool lastIsPrintable( false );

 char twoLetterC2nd;
 t_int16 refCode;
 t_gUniPage ucCompat;
 t_gUniPage ucAlt;
 int letterUp1Down2;

 ASSERTION(fIn,"fIn");
 ASSERTION(fOut,"fOut");

 DBGPRINT("DBG: go_file maxChr=%u, doViewAll?%c\n",
	  maxChr,
	  ISyORn(opt.doViewAll));

 if ( maxChr<256 ) return 8;

 for ( ; (error = fscanf(fIn,"%c",&chr)!=1)==0; ) {
     unChr = (unsigned)chr;
     bool isSymbol;
     bool isPrintable( (uc.mapTab[ unChr ].extUtMask & EXTUMASK_ISPRINT)!=0 ||
		       (opt.doExtended==true) );

     if ( chr=='\r' && opt.isOutDOS ) continue;

     if ( opt.doSkip8Bit==true && unChr>=127 ) {
	 lastIsPrintable = false;
	 continue;
     }

     refCode = unCode.refCode[ unChr ];
     refChrUse = unCode.refChrUse[ unChr ];
     ASSERTION(refCode>=0,"refCode>=0");
     DBGPRINT_MIN("\tASCII:%4u refCode: %d\t%c\n\n",unChr,refCode,chr);

     if ( opt.doViewAll ) {
	 gString* name = uc.mapTab[ unChr ].symbolName;
	 char* symNameStr = name ? name->Str() : nil;
	 // View all printable chars
	 //if ( unChr > 128 ) {
	 //    printf("\nDBG: %ud '%c' isPrintable?%c\n",
	 //	    unChr,
	 //	    chr,
	 //	    ISyORn(isPrintable));
	 //}
	 if ( isPrintable ) {
	     fprintf(fOut,"%c",chr);
	 }
	 else {
	     fprintf(fOut,"[[%s]]%s",
		     symNameStr,
		     strg_nl( opt.isDumpLinear==false, opt ));
	 }
	 continue;
     }

     if ( refCode<=0 ) {
	 lastIsPrintable = false;
	 continue;
     }

     ucCompat = uc.mapTab[ unChr ].ucCompat;

     twoLetterC2nd = uc.mapTab[ unChr ].twoLetterCat[ 1 ];
     if ( uc.mapTab[ unChr ].twoLetterCat[ 0 ]=='L' )
	 letterUp1Down2 = (twoLetterC2nd=='u') + ((twoLetterC2nd=='l')*2);
     else
	 letterUp1Down2 = 0;

     isSymbol = twoLetterC2nd=='k';  // On Unicode this is marked as '...;Sk;...'

     switch ( opt.caseKind ) {
     case 0:
	 if ( refChrUse==0 ) {
	     lastIsPrintable = false;
	     continue;
	 }
	 chr = refChrUse;
	 break;

     case 1:
	 ucAlt = uc.mapTab[ unChr ].ucUpper;
	 if ( ucAlt )
	     chr = (char)ucAlt;
	 else
	     if ( letterUp1Down2==2 )
		 chr = '?';
	 break;

     case 2:
	 ucAlt = uc.mapTab[ unChr ].ucDown;
	 if ( ucAlt )
	     chr = (char)ucAlt;
	 break;

     case 8:
	 chr = (t_uchar)refCode;
	 // Apparently similar to:
	 //	char chrCompat( ucCompat );
	 //	chr = chrCompat ? chrCompat : refChrUse;
	 break;

     default:
	 ASSERTION_FALSE("Uops");
     }

     bool lastPrintOrNew( opt.isDumpLinear ?
			  true :
			  (lastIsPrintable==true || lastChr=='\n') );
     if ( opt.isOutDOS )
	 fprintf(fOut,"%s%s%c",
		 strg_nl( lastPrintOrNew==false, opt ),
		 chr=='\n' ? "\r" : "\0",
		 chr);
     else
	 fprintf(fOut,"%s%c",
		 strg_nl( lastPrintOrNew==false, opt ),
		 chr);
     lastIsPrintable = true;
     lastChr = chr;
 }

 return 0;
}
////////////////////////////////////////////////////////////
int go_cmd (gArg& arg,
	    FILE* fOut,
	    FILE* fRepErr,
	    sOptStrgs& opt,
	    int aCmd)
{
 int error = 0, thisError, countIter = 0;
 char* str;
 unsigned i, n = arg.N();
 FILE* fIn = nil;

 ASSERTION(fRepErr,"fRepErr");
 ASSERTION(fOut,"fOut");
 ASSERTION(ptrUcBin,"ptrUcBin");

 for (i=1; i<=n; i++) {
     str = arg.Str( i );
     if ( str[0]==0 ) continue;
     if ( opt.isVeryVerbose || (opt.isVerbose==true && n>1) ) fprintf(stderr,":::%s\n",str);
     countIter++;
     fIn = fopen( str, "rb" );
     thisError = fIn ? 0 : 2;
     if ( thisError==0 ) {
	 thisError = go_file( fIn,
			      fOut,
			      fRepErr,
			      *ptrUcBin, *ptrUncode, opt );
	 fclose( fIn );
     }
     if ( thisError!=0 ) {
	 fprintf(fRepErr,"Error (%d) handling: %s\n",thisError,str);
	 if ( error==0 ) error = thisError;
     }
     DBGPRINT("DBG: 'go_file' returned %d count=%d\n",thisError,countIter);
 }

 if ( countIter==0 ) {
     error = go_file( stdin, fOut, fRepErr, *ptrUcBin, *ptrUncode, opt );
 }

 return error;
}
////////////////////////////////////////////////////////////
int run_args (char** argv, char** envp, int& result)
{
 int error, thisError = -1;
 gArg arg( argv, envp );
 sOptStrgs opt;

 unsigned n;
 FILE* fOut = stdout;
 FILE* fRepErr = stderr;

 result = 0;

 arg.AddParams( 1, "-vv|--verbose -h|--help --version\
 -c:%s|--case:%s\
 -o:%s|--output:%s\
 -8|--skip-8bit\
 -e|--extended\
 --dos-mode\
 --dump-linear\
 --view\
" );

 error = arg.FlushParams();
 n = arg.N();
 arg.FindOptionOccurr( "verbose", opt.isVerbose, opt.isVeryVerbose );

 if ( arg.FindOption('h') || error!=0 ) {
     print_help( arg.Program() );
     return 0;
 }
 if ( arg.FindOption("version") ) {
     print_version( arg.Program() );
     return 0;
 }

 opt.isOutDOS = arg.FindOption("dos-mode");
 opt.isDumpLinear = arg.FindOption("dump-linear");
 opt.doSkip8Bit = arg.FindOption('8');
 opt.doExtended = arg.FindOption('e');
 opt.doViewAll = arg.FindOption("view");

 if ( opt.doSkip8Bit && opt.doExtended ) {
     fprintf(stderr,"-8 and -e cannot be used on the same time.\n");
     return 1;
 }

 if ( arg.FindOption('c',opt.sCase) && opt.sCase.IsEmpty()==false ) {
     gString s( opt.sCase );
     int caseKind( 0 );
     caseKind += s.Match( "up", true ) || s.Match( "upper", true );
     caseKind += (s.Match( "down", true ) || s.Match( "lower", true ))*2;
     caseKind += s.Match( "normal", true )*8;
     if ( caseKind==0 ) {
	 fprintf(stderr,"Invalid case specified: either normal, 'up'/'upper' or 'down'/'lower' can be chosen.\n");
	 return 1;
     }
     opt.caseKind = caseKind;
 }

 if ( arg.FindOption('o',opt.sDumpFile) && opt.sDumpFile.IsEmpty()==false ) {
     fOut = fopen( opt.sDumpFile.Str(), opt.isOutDOS ? "wt" : "wb" );
     if ( fOut==nil ) {
	 fprintf(stderr,"Unable to write file: %s\n",
		 opt.sDumpFile.Str());
	 return 4;
     }
 }

 // parsing nearly ends...
 if ( opt.caseKind>0 && opt.doViewAll ) {
     opt.caseKind = 0;
     if ( opt.isVerbose )
	 fprintf(fRepErr,"Warning: --view and -c options specified; ignoring -c (%s)\n",opt.sCase.Str());
 }
 // .

 thisError = unicode_leveling_latin1();
 DBGPRINT("DBG: unicode_leveling_latin1 returned %d\n",thisError);

 // go-ahead:
 error = go_cmd( arg, fOut, fRepErr, opt, 0 );
 DBGPRINT("DBG: go_cmd returned %d\n",error);

 return result = error;
}

int go (char** argv, char** envp, int& result)
{
 int error;
 FILE* fRepErr( stderr );

 ptrUncode = new gOptUnCode;
 ptrUcBin = new sUnCodeBin;
 if ( ptrUncode==nil || ptrUcBin==nil ) return nil;

 UNI_STG_ADJUST;
 error = unicode_read();
 result = error ? 4 : 0;

 char* pathStr( ptrUncode->UniThbFileStr() );
 if ( error ) {
     fprintf(fRepErr,"Unable to read Unicode-plus THB: %s, error: %d\n",
	     pathStr,
	     result);
     return error;
 }

 error = run_args( argv, envp, result );

 delete ptrUncode; ptrUncode = nil;
 delete ptrUcBin; ptrUcBin = nil;

 return error;
}
////////////////////////////////////////////////////////////
int main (int argc, char** argv, char** envp)
{
 int error, result( 0 );

 gINIT;

 error = go( argv, envp, result );

 gEND;

 DBGPRINT_MIN("go: %d, result=%d\n",error,result);

 return result;
}

