// gSupStrUni.cpp -- Version 0.1

// Just an easier interface with libgobj UniCode support

#include <string.h>
#include "gcontrol.h"
#include "gfilestat.h"

#include "gSupStrUni.h"
////////////////////////////////////////////////////////////
gOptUnCode::gOptUnCode (char* installedPrefixThb, unsigned maxUniPathLenAdd)
    : nRefsInvalid( -1 ),
      sUniTblPathBuf( (unsigned)strlen(UNI_CODE_DEF_PATH) + maxUniPathLenAdd, '\0' )
{
 lastOpError = internalPathBuilder( installedPrefixThb, "unicode256.thb" );
 DBGPRINT("DBG: internalPathBuilder returned %d\n",lastOpError);
 // Note: nRefsInvalid -1 means no leveling was done
}

char* gOptUnCode::UniThbFileStr ()
{
 return gStrControl::Self().GetPathNameTHB();
}

int gOptUnCode::internalPathBuilder (char* installedPrefixThb,
				     char* defName)
{
 char* buf = sUniTblPathBuf.Str();
 char* emptyStr = (char*)UNI_DEF_INSTALLER_PREFIX;
 unsigned len;

 ASSERTION(defName,"defName");
 ASSERTION(defName[0],"defName[0]");

 if ( installedPrefixThb==nil ) {
     installedPrefixThb = getenv( "APPDATA" );
 }
 if ( installedPrefixThb==nil )
     installedPrefixThb = emptyStr;

 sprintf( buf, UNI_CODE_DEF_PATH, installedPrefixThb );
 DBGPRINT("DBG: sUniTblPathBuf: '%s'\n",buf);

 sUniTblPath.Set( buf );
 len = sUniTblPath.Length();
 if ( sUniTblPath[ len ]==gSLASHCHR ) {
     sUniTblPath[ len ] = '\0';
 }

 // Set the name, shouldn't have any slash char
 sUniTblName.Set( defName );
 len = sUniTblName.Find( gSLASHCHR );
 if ( len>0 ) return -1;

 return 0;
}

int gOptUnCode::CheckUniTHB ()
{
 gFileStat dirTest( gStrControl::Self().GetPathUnicode() );
 if ( dirTest.IsOk()==false || dirTest.IsDirectory()==false )
     return 3;

 gFileStat fileTest( UniThbFileStr() );

 return fileTest.IsOk() ? 0 : 2;
}
////////////////////////////////////////////////////////////
// Functions
////////////////////////////////////////////////////////////
int unicode_read ()
{
 ASSERTION(ptrUncode,"ptrUncode");
 ASSERTION(ptrUcBin,"ptrUcBin");
 return ucbin_read_opt( *ptrUncode, *ptrUcBin );
}

int unicode_leveling_latin1 ()
{
 return unicode_leveling( UNI_PART_LATIN1 );
}

int unicode_leveling (t_unipart part)
{
 ASSERTION(ptrUncode,"ptrUncode");
 ASSERTION(ptrUcBin,"ptrUcBin");
 return ucbin_leveling( part, true, *ptrUncode, *ptrUcBin );
}
////////////////////////////////////////////////////////////
int ucbin_read (char* thbFilename, sUnCodeBin& ucBin)
{
 if ( thbFilename==nil ) return -1;
 return ucBin.ReadFromFile( thbFilename );
}

int ucbin_read_opt (gOptUnCode& unCode, sUnCodeBin& ucBin)
{
 int error = unCode.lastOpError;

 DBGPRINT_MIN("DBG: ucbin_read_opt: unCode error=%d\n",error);
 if ( error ) return -1;
 error = ucbin_read( unCode.UniThbFileStr(), ucBin );
 DBGPRINT_MIN("DBG: read_ucbin (%s) returned: %d\n",
	      unCode.UniThbFile().Str(),
	      error);
 return error!=0;
}

int ucbin_leveling (t_unipart part, bool doThbLeveling, gOptUnCode& unCode, sUnCodeBin& ucBin)
{
 unsigned uIdx, n( ucBin.n );
 t_uchar chr( '\0' ), two, refChrUse;
 t_gUniPage thisUcCompat;

 DBGPRINT("DBG: ucbin_leveling: part=%u, n=%u\n",(unsigned)part,n);
 if ( n<256 ) return 8;  // error-code 8: ucbin too short
 unCode.nRefsInvalid = 0;

 for (uIdx=0; uIdx<0x20; uIdx++, chr++) {
     unCode.refCode[ uIdx ] = (t_int16)(unCode.refChrUse[ uIdx ] = (chr=='\t' || chr=='\n') ? chr : 0);
 }

 // Run from blank (ASCII 32d) till 126d
 for ( ; uIdx<127; uIdx++, chr++) {
     two = ucBin.mapTab[ uIdx ].twoLetterCat[ 0 ];
     if ( doThbLeveling ) {
	 if ( two=='L' )
	     ucBin.mapTab[ uIdx ].ucCompat = chr;  // Here, a trivial leveling!
     }
     thisUcCompat = ucBin.mapTab[ uIdx ].ucCompat;
     unCode.refCode[ uIdx ] = (t_int16)(unCode.refChrUse[ uIdx ] = chr);
     DBGPRINT_MIN("%3u 0x%02X has UniCode=0x%04X\t%c %3u '%c' %s\n",
		  uIdx,
		  uIdx,
		  ucBin.mapTab[uIdx].uniCode,
		  (uIdx >= 127 && uIdx < 160) ? '.' : chr,
		  thisUcCompat,
		  (thisUcCompat > ' ' && thisUcCompat < 127) ? thisUcCompat : '*',
		  two=='L' ? ucBin.mapTab[uIdx].symbolName->Str() : "\0");
 }
 unCode.refCode[ uIdx ] = unCode.refChrUse[ uIdx ] = 0;  // 127d is the escape character, never printable!

 // sequential numbering 128..255
 for (uIdx=128; uIdx<256 && uIdx<n; uIdx++) {
     bool isPrintable( (ucBin.mapTab[ uIdx ].extUtMask & UNI_PRINTABLE_LATIN)!=0 );
     two = ucBin.mapTab[ uIdx ].twoLetterCat[ 0 ];
     thisUcCompat = ucBin.mapTab[ uIdx ].ucCompat;
     bool isValidHere( thisUcCompat!=0 );
     refChrUse = (isPrintable && isValidHere) ? (t_uchar)uIdx : 0;
     unCode.refCode[ uIdx ] = thisUcCompat > MAX_INT16_I ? 0 : (t_int16)thisUcCompat;
     unCode.refChrUse[ uIdx ] = refChrUse;
     DBGPRINT("%3u 0x%02X UniCode=0x%04X\t%c %3u '%c','%c' (%03u) : {%-2s} Mask=0x%X %s\n",
	      uIdx,
	      uIdx,
	      ucBin.mapTab[uIdx].uniCode,
	      (uIdx >= 127 && uIdx < 160) ? '.' : uIdx,
	      thisUcCompat,
	      (thisUcCompat > ' ' && thisUcCompat < 127) ? thisUcCompat : '*',
	      refChrUse ? refChrUse : '?',
	      unCode.refCode[ uIdx ],
	      ucBin.mapTab[ uIdx ].thrLetterStyle,
	      ucBin.mapTab[ uIdx ].extUtMask,
	      ucBin.mapTab[uIdx].symbolName->Str());
     unCode.nRefsInvalid += uIdx!=ucBin.mapTab[uIdx].uniCode;
 }
 for ( ; uIdx<256; uIdx++) {
     unCode.refCode[ uIdx ] = -1;
     unCode.refChrUse[ uIdx ] = 0;
 }

 return 0;
}
////////////////////////////////////////////////////////////

